from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.report import Report
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import MessageRole, PendingActionType, RunStatus, WorkforceMode
from blackgeorge.desk import Desk
from blackgeorge.logging import StructuredLogger, get_logger
from blackgeorge.multimodal import encode_file
from blackgeorge.session import WorkerSession
from blackgeorge.tools import Tool, Toolbelt, Toolkit, agenerate_image, generate_image, tool
from blackgeorge.worker import Worker
from blackgeorge.workforce import Workforce

Brief = Job
RunOutput = Report

__all__ = [
    "Brief",
    "Desk",
    "Event",
    "Job",
    "Message",
    "MessageRole",
    "PendingAction",
    "PendingActionType",
    "Report",
    "RunOutput",
    "RunStatus",
    "StructuredLogger",
    "Tool",
    "ToolCall",
    "Toolbelt",
    "Toolkit",
    "Workforce",
    "WorkforceMode",
    "Worker",
    "WorkerSession",
    "agenerate_image",
    "encode_file",
    "generate_image",
    "get_logger",
    "tool",
]



import json
from collections.abc import Mapping, Sequence
from typing import cast

import litellm

from blackgeorge.tools.base import ToolResult
from blackgeorge.tools.decorators import tool


def _value(obj: object, key: str) -> object | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, object], obj).get(key)
    return getattr(obj, key, None)


def _extract_image_generation_response(response: object) -> dict[str, str]:
    data = _value(response, "data")
    if not isinstance(data, Sequence) or isinstance(data, str | bytes):
        return {}
    if len(data) == 0:
        return {}

    image = data[0]
    result: dict[str, str] = {}

    url = _value(image, "url")
    if isinstance(url, str) and url:
        result["url"] = url

    b64_json = _value(image, "b64_json")
    if isinstance(b64_json, str) and b64_json:
        result["b64_json"] = b64_json

    revised_prompt = _value(image, "revised_prompt")
    if isinstance(revised_prompt, str) and revised_prompt:
        result["revised_prompt"] = revised_prompt

    return result


def _extract_image_url_from_content_blocks(content: object) -> str | None:
    if not isinstance(content, Sequence) or isinstance(content, str | bytes):
        return None
    for block in content:
        block_type = _value(block, "type")
        image_url = _value(block, "image_url")
        if (
            block_type not in {"image_url", "input_image", "output_image"}
            and image_url is None
            and _value(block, "url") is None
        ):
            continue
        if isinstance(image_url, str) and image_url:
            return image_url
        if image_url is not None:
            nested_url = _value(image_url, "url")
            if isinstance(nested_url, str) and nested_url:
                return nested_url
        direct_url = _value(block, "url")
        if isinstance(direct_url, str) and direct_url:
            return direct_url
    return None


def _extract_chat_image_response(response: object) -> dict[str, str]:
    choices = _value(response, "choices")
    if not isinstance(choices, Sequence) or isinstance(choices, str | bytes):
        return {}
    if len(choices) == 0:
        return {}

    message = _value(choices[0], "message")
    images = _value(message, "images")
    result: dict[str, str] = {}

    if isinstance(images, Sequence) and not isinstance(images, str | bytes) and len(images) > 0:
        first_image = images[0]
        image_url = _value(first_image, "image_url")
        url: object | None = None
        if image_url is not None:
            url = _value(image_url, "url")
        if url is None:
            url = _value(first_image, "url")
        if isinstance(url, str) and url:
            result["url"] = url

    content = _value(message, "content")
    if "url" not in result:
        if isinstance(content, str) and content.startswith("data:image/"):
            result["url"] = content
        else:
            block_url = _extract_image_url_from_content_blocks(content)
            if isinstance(block_url, str) and block_url:
                result["url"] = block_url

    return result


def _generate_via_chat_completion(prompt: str, model: str) -> dict[str, str]:
    response = litellm.completion(
        model=model,
        messages=[{"role": "user", "content": prompt}],
        modalities=["image", "text"],
    )
    return _extract_chat_image_response(response)


async def _agenerate_via_chat_completion(prompt: str, model: str) -> dict[str, str]:
    response = await litellm.acompletion(
        model=model,
        messages=[{"role": "user", "content": prompt}],
        modalities=["image", "text"],
    )
    return _extract_chat_image_response(response)


def _tool_result(payload: dict[str, str]) -> ToolResult:
    compact = dict(payload)
    url = compact.get("url")
    if isinstance(url, str) and url.startswith("data:image/"):
        compact["url"] = "[data-url omitted]"
    b64_json = compact.get("b64_json")
    if isinstance(b64_json, str) and b64_json:
        compact["b64_json"] = "[b64_json omitted]"
    return ToolResult(
        content=json.dumps(compact, ensure_ascii=True, default=str),
        data=payload,
    )


@tool(description="Generate an image from a text prompt using an AI model")
def generate_image(
    prompt: str,
    model: str = "openrouter/google/gemini-3-pro-image-preview",
    size: str = "1024x1024",
    quality: str = "standard",
) -> ToolResult:
    generation_error: Exception | None = None
    try:
        response = litellm.image_generation(
            prompt=prompt,
            model=model,
            size=size,
            quality=quality,
        )
        result = _extract_image_generation_response(response)
        if result:
            return _tool_result(result)
    except Exception as exc:
        generation_error = exc

    result = _generate_via_chat_completion(prompt=prompt, model=model)
    if result:
        return _tool_result(result)
    if generation_error is not None:
        raise generation_error
    return _tool_result({})


@tool(description="Generate an image asynchronously from a text prompt")
async def agenerate_image(
    prompt: str,
    model: str = "openrouter/google/gemini-3-pro-image-preview",
    size: str = "1024x1024",
    quality: str = "standard",
) -> ToolResult:
    generation_error: Exception | None = None
    try:
        response = await litellm.aimage_generation(
            prompt=prompt,
            model=model,
            size=size,
            quality=quality,
        )
        result = _extract_image_generation_response(response)
        if result:
            return _tool_result(result)
    except Exception as exc:
        generation_error = exc

    result = await _agenerate_via_chat_completion(prompt=prompt, model=model)
    if result:
        return _tool_result(result)
    if generation_error is not None:
        raise generation_error
    return _tool_result({})



import asyncio
import atexit
import json
import warnings
from collections.abc import AsyncIterator, Callable, Iterator
from typing import Any, cast

import litellm
from pydantic import BaseModel, TypeAdapter

from blackgeorge.adapters.base import BaseModelAdapter, ModelResponse
from blackgeorge.adapters.instructor_client import instructor_clients
from blackgeorge.adapters.litellm_callbacks import (
    _callback_context,
    emit_llm_completed,
    emit_llm_failed,
    emit_llm_started,
)
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.utils import new_id

_litellm_runtime_configured = False


def _close_coroutine(coroutine: Any) -> None:
    close = getattr(coroutine, "close", None)
    if callable(close):
        close()


def _safe_litellm_shutdown() -> None:
    try:
        from litellm.litellm_core_utils.logging_worker import GLOBAL_LOGGING_WORKER

        flush = getattr(GLOBAL_LOGGING_WORKER, "_flush_on_exit", None)
        if callable(flush):
            flush()
    except Exception:
        pass
    try:
        from litellm.llms.custom_httpx.async_client_cleanup import close_litellm_async_clients
    except Exception:
        return
    cleanup_coroutine = close_litellm_async_clients()
    try:
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(cleanup_coroutine)
            cleanup_coroutine = None
        finally:
            loop.close()
    except Exception:
        pass
    finally:
        if cleanup_coroutine is not None:
            _close_coroutine(cleanup_coroutine)


def _patch_logging_worker_enqueue() -> None:
    try:
        from litellm.litellm_core_utils.logging_worker import GLOBAL_LOGGING_WORKER
    except Exception:
        return
    worker = cast(Any, GLOBAL_LOGGING_WORKER)
    if getattr(worker, "_blackgeorge_enqueue_patch", False):
        return
    original_ensure_initialized_and_enqueue = worker.ensure_initialized_and_enqueue

    def ensure_initialized_and_enqueue(async_coroutine: Any) -> None:
        original_ensure_initialized_and_enqueue(async_coroutine)
        if getattr(worker, "_queue", None) is None:
            _close_coroutine(async_coroutine)

    worker.ensure_initialized_and_enqueue = ensure_initialized_and_enqueue
    worker._blackgeorge_enqueue_patch = True


def _configure_litellm_runtime() -> None:
    global _litellm_runtime_configured
    if _litellm_runtime_configured:
        return
    litellm.disable_streaming_logging = True
    if hasattr(litellm, "_async_client_cleanup_registered"):
        litellm._async_client_cleanup_registered = True
    _patch_logging_worker_enqueue()
    atexit.register(_safe_litellm_shutdown)
    _litellm_runtime_configured = True


def _get(obj: Any, key: str, default: Any = None) -> Any:
    if isinstance(obj, dict):
        return obj.get(key, default)
    return getattr(obj, key, default)


def _response_content(response: Any) -> str | None:
    choices = _get(response, "choices", [])
    message = _get(choices[0], "message") if choices else None
    return _get(message, "content") if message else None


def _build_json_schema(response_schema: Any) -> dict[str, Any] | None:
    if isinstance(response_schema, TypeAdapter):
        try:
            return response_schema.json_schema()
        except Exception:
            return None
    if isinstance(response_schema, type) and issubclass(response_schema, BaseModel):
        try:
            return response_schema.model_json_schema()
        except Exception:
            return None
    try:
        adapter = TypeAdapter(response_schema)
    except Exception:
        return None
    try:
        return adapter.json_schema()
    except Exception:
        return None


def _response_format(response_schema: Any) -> dict[str, Any] | None:
    schema = _build_json_schema(response_schema)
    if schema is None:
        return None
    name = getattr(response_schema, "__name__", None)
    if name is None:
        inner = getattr(response_schema, "_type", None)
        name = getattr(inner, "__name__", "response_schema")
    return {"type": "json_schema", "json_schema": {"name": name, "schema": schema, "strict": True}}


def _parse_structured_json(response_schema: Any, content: str) -> Any:
    if isinstance(response_schema, TypeAdapter):
        return response_schema.validate_json(content)
    if isinstance(response_schema, type) and issubclass(response_schema, BaseModel):
        return response_schema.model_validate_json(content)
    adapter = TypeAdapter(response_schema)
    return adapter.validate_json(content)


def _structured_retry_prompt(error: Exception) -> str:
    return (
        "Fix validation errors and return only JSON that matches the schema. "
        "Do not output YAML. Do not wrap objects as strings. "
        "If a field is a list, each item must be a JSON object, not a string. "
        f"Validation errors: {error}"
    )


def _is_base_model_schema(response_schema: Any) -> bool:
    return isinstance(response_schema, type) and issubclass(response_schema, BaseModel)


def _structured_json_retry(
    *,
    model: str,
    payload: list[dict[str, Any]],
    response_schema: Any,
    retries: int,
    response_format: dict[str, Any] | None,
) -> Any:
    attempts = 0
    while True:
        try:
            if response_format is None:
                response = litellm.completion(model=model, messages=payload)
            else:
                response = litellm.completion(
                    model=model,
                    messages=payload,
                    response_format=response_format,
                )
            content = _response_content(response)
            if not content:
                raise ValueError("Empty response content")
            return _parse_structured_json(response_schema, content)
        except Exception as exc:
            if attempts >= retries:
                raise
            payload.append({"role": "user", "content": _structured_retry_prompt(exc)})
            attempts += 1


async def _astructured_json_retry(
    *,
    model: str,
    payload: list[dict[str, Any]],
    response_schema: Any,
    retries: int,
    response_format: dict[str, Any] | None,
) -> Any:
    attempts = 0
    while True:
        try:
            if response_format is None:
                response = await litellm.acompletion(model=model, messages=payload)
            else:
                response = await litellm.acompletion(
                    model=model,
                    messages=payload,
                    response_format=response_format,
                )
            content = _response_content(response)
            if not content:
                raise ValueError("Empty response content")
            return _parse_structured_json(response_schema, content)
        except Exception as exc:
            if attempts >= retries:
                raise
            payload.append({"role": "user", "content": _structured_retry_prompt(exc)})
            attempts += 1


def _parse_tool_calls(message: Any) -> list[ToolCall]:
    tool_calls = _get(message, "tool_calls", []) or []
    if not isinstance(tool_calls, list):
        return []
    parsed: list[ToolCall] = []

    for call in tool_calls:
        call_like = isinstance(call, dict) or hasattr(call, "function") or hasattr(call, "id")
        if not call_like:
            continue
        function = _get(call, "function", {})
        function_like = (
            isinstance(function, dict)
            or hasattr(function, "name")
            or hasattr(function, "arguments")
        )
        if not function_like:
            function = {}
        name_raw = _get(function, "name")
        name = name_raw.strip() if isinstance(name_raw, str) else ""
        arguments_raw = _get(function, "arguments")
        arguments: dict[str, Any] = {}
        error: str | None = None
        if not name:
            error = "Missing tool name"

        if isinstance(arguments_raw, dict):
            arguments = arguments_raw
        elif isinstance(arguments_raw, str):
            if arguments_raw:
                try:
                    arguments = json.loads(arguments_raw)
                except json.JSONDecodeError as e:
                    message_text = (
                        f"Invalid JSON in tool arguments: {e}. Raw: {arguments_raw[:100]}"
                    )
                    error = f"{error}; {message_text}" if error else message_text
                    arguments = {}
        elif arguments_raw is not None:
            message_text = f"Unsupported tool arguments type: {type(arguments_raw).__name__}"
            error = f"{error}; {message_text}" if error else message_text

        call_id_raw = _get(call, "id")
        call_id = call_id_raw if isinstance(call_id_raw, str) and call_id_raw else new_id()
        parsed.append(ToolCall(id=call_id, name=name, arguments=arguments, error=error))

    return parsed


def _supports_parallel_function_calling(model: str) -> bool:
    model_cost = getattr(litellm, "model_cost", {})
    if not isinstance(model_cost, dict):
        return False
    candidates = [model]
    model_parts = model.split("/")
    if len(model_parts) > 1:
        for index in range(1, len(model_parts)):
            candidates.append("/".join(model_parts[index:]))
    for candidate in candidates:
        info = model_cost.get(candidate)
        if not isinstance(info, dict):
            continue
        supported = info.get("supports_parallel_function_calling")
        if isinstance(supported, bool):
            return supported
    return False


def _parse_response(response: Any) -> ModelResponse:
    choices = _get(response, "choices", [])
    message = _get(choices[0], "message") if choices else None
    content = _get(message, "content") if message else None
    reasoning_content = _get(message, "reasoning_content") if message else None
    thinking_blocks = _get(message, "thinking_blocks") if message else None
    tool_calls = _parse_tool_calls(message) if message else []
    usage = _get(response, "usage", {}) or {}
    if isinstance(usage, BaseModel):
        usage = usage.model_dump(mode="json", warnings=False)
    return ModelResponse(
        content=content,
        reasoning_content=reasoning_content,
        thinking_blocks=thinking_blocks,
        tool_calls=tool_calls,
        usage=usage,
        raw=response,
    )


def _build_litellm_params(
    *,
    model: str,
    messages: list[dict[str, Any]],
    tools: list[dict[str, Any]] | None,
    tool_choice: str | dict[str, Any] | None,
    temperature: float | None,
    max_tokens: int | None,
    stream: bool,
    stream_options: dict[str, Any] | None,
    thinking: dict[str, Any] | None,
    drop_params: bool | None,
    extra_body: dict[str, Any] | None,
) -> dict[str, Any]:
    params: dict[str, Any] = {
        "model": model,
        "messages": messages,
        "stream": stream,
    }
    if tools:
        params["tools"] = tools
        if tool_choice is not None:
            params["tool_choice"] = tool_choice
    if temperature is not None:
        params["temperature"] = temperature
    if max_tokens is not None:
        params["max_tokens"] = max_tokens
    if stream and stream_options is not None:
        params["stream_options"] = stream_options
    if thinking is not None:
        params["thinking"] = thinking
    if drop_params is not None:
        params["drop_params"] = drop_params
    if extra_body is not None:
        params["extra_body"] = extra_body
    if tools and _supports_parallel_function_calling(model):
        params["parallel_tool_calls"] = True
    return params


def _stream_usage(chunk: Any) -> dict[str, Any] | None:
    usage = _get(chunk, "usage")
    if isinstance(usage, dict):
        return usage
    if isinstance(usage, BaseModel):
        return usage.model_dump(mode="json", warnings=False)
    return None


class _SyncStreamWithEvents:
    def __init__(self, model: str, stream: Iterator[Any]) -> None:
        self._model = model
        self._stream = stream
        self._usage: dict[str, Any] = {}
        self._closed = False

    def __iter__(self) -> Iterator[Any]:
        try:
            for chunk in self._stream:
                usage = _stream_usage(chunk)
                if usage:
                    self._usage = usage
                yield chunk
        except Exception as exc:
            self._emit_failed(exc)
            raise
        self._emit_completed()

    def close(self) -> None:
        if self._closed:
            return
        try:
            close_fn = getattr(self._stream, "close", None)
            if callable(close_fn):
                close_fn()
        except Exception as exc:
            self._emit_failed(exc)
            raise
        self._emit_completed()

    def _emit_completed(self) -> None:
        if self._closed:
            return
        self._closed = True
        emit_llm_completed(self._model, {"usage": self._usage})

    def _emit_failed(self, exc: Exception) -> None:
        if self._closed:
            return
        self._closed = True
        emit_llm_failed(self._model, exc)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._stream, name)


class _AsyncStreamWithEvents:
    def __init__(self, model: str, stream: AsyncIterator[Any]) -> None:
        self._model = model
        self._stream = stream
        self._iterator = stream.__aiter__()
        self._usage: dict[str, Any] = {}
        self._closed = False

    def __aiter__(self) -> "_AsyncStreamWithEvents":
        return self

    async def __anext__(self) -> Any:
        if self._closed:
            raise StopAsyncIteration
        try:
            chunk = await self._iterator.__anext__()
        except StopAsyncIteration:
            self._emit_completed()
            raise
        except Exception as exc:
            self._emit_failed(exc)
            raise
        usage = _stream_usage(chunk)
        if usage:
            self._usage = usage
        return chunk

    async def aclose(self) -> None:
        if self._closed:
            return
        try:
            aclose_fn = getattr(self._stream, "aclose", None)
            if callable(aclose_fn):
                await aclose_fn()
            else:
                close_fn = getattr(self._stream, "close", None)
                if callable(close_fn):
                    close_fn()
        except Exception as exc:
            self._emit_failed(exc)
            raise
        self._emit_completed()

    def _emit_completed(self) -> None:
        if self._closed:
            return
        self._closed = True
        emit_llm_completed(self._model, {"usage": self._usage})

    def _emit_failed(self, exc: Exception) -> None:
        if self._closed:
            return
        self._closed = True
        emit_llm_failed(self._model, exc)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._stream, name)


def _wrap_stream_with_events(model: str, response: Any, *, prefer_async: bool) -> Any:
    if prefer_async and hasattr(response, "__aiter__"):
        return _AsyncStreamWithEvents(model, cast(AsyncIterator[Any], response))
    if hasattr(response, "__iter__"):
        return _SyncStreamWithEvents(model, cast(Iterator[Any], response))
    if hasattr(response, "__aiter__"):
        return _AsyncStreamWithEvents(model, cast(AsyncIterator[Any], response))
    return response


class LiteLLMAdapter(BaseModelAdapter):
    def __init__(self) -> None:
        warnings.filterwarnings("ignore", message="Pydantic serializer warnings")
        _configure_litellm_runtime()

    def set_callback_context(
        self, run_id: str, emit: Callable[[str, str, dict[str, Any]], None]
    ) -> None:
        _callback_context.set({"run_id": run_id, "emit": emit})

    def clear_callback_context(self) -> None:
        _callback_context.set(None)

    def complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        tool_choice: str | dict[str, Any] | None,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse | list[dict[str, Any]]:
        litellm_params = _build_litellm_params(
            model=model,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            thinking=thinking,
            drop_params=drop_params,
            extra_body=extra_body,
        )

        emit_llm_started(model, len(messages), len(tools) if tools else 0)
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Pydantic serializer warnings")
                response = litellm.completion(**litellm_params)
            if stream:
                return cast(
                    list[dict[str, Any]],
                    _wrap_stream_with_events(model, response, prefer_async=False),
                )
            emit_llm_completed(model, response)
            return _parse_response(response)
        except Exception as exc:
            emit_llm_failed(model, exc)
            raise

    async def acomplete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        tools: list[dict[str, Any]] | None,
        tool_choice: str | dict[str, Any] | None,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
    ) -> ModelResponse | Any:
        litellm_params = _build_litellm_params(
            model=model,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            thinking=thinking,
            drop_params=drop_params,
            extra_body=extra_body,
        )

        emit_llm_started(model, len(messages), len(tools) if tools else 0)
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Pydantic serializer warnings")
                response = await litellm.acompletion(**litellm_params)
            if stream:
                return _wrap_stream_with_events(model, response, prefer_async=True)
            emit_llm_completed(model, response)
            return _parse_response(response)
        except Exception as exc:
            emit_llm_failed(model, exc)
            raise

    def structured_complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = list(messages)
        response_format = _response_format(response_schema)
        response_format_accepted = False
        if response_format is not None:
            try:
                response = litellm.completion(
                    model=model,
                    messages=payload,
                    response_format=response_format,
                )
                response_format_accepted = True
                content = _response_content(response)
                if content:
                    return _parse_structured_json(response_schema, content)
            except Exception:
                response_format_accepted = False
        retries = max(retries, 3)
        if not _is_base_model_schema(response_schema):
            manual_response_format = response_format if response_format_accepted else None
            return _structured_json_retry(
                model=model,
                payload=payload,
                response_schema=response_schema,
                retries=retries,
                response_format=manual_response_format,
            )
        client = instructor_clients.get(model, async_client=False)
        attempts = 0
        while True:
            try:
                return client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": _structured_retry_prompt(exc),
                    }
                )
                attempts += 1

    async def astructured_complete(
        self,
        *,
        model: str,
        messages: list[dict[str, Any]],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = list(messages)
        response_format = _response_format(response_schema)
        response_format_accepted = False
        if response_format is not None:
            try:
                response = await litellm.acompletion(
                    model=model,
                    messages=payload,
                    response_format=response_format,
                )
                response_format_accepted = True
                content = _response_content(response)
                if content:
                    return _parse_structured_json(response_schema, content)
            except Exception:
                response_format_accepted = False
        retries = max(retries, 3)
        if not _is_base_model_schema(response_schema):
            manual_response_format = response_format if response_format_accepted else None
            return await _astructured_json_retry(
                model=model,
                payload=payload,
                response_schema=response_schema,
                retries=retries,
                response_format=manual_response_format,
            )
        client = instructor_clients.get(model, async_client=True)
        attempts = 0
        while True:
            try:
                return await client.chat.completions.create(
                    model=model,
                    messages=payload,
                    response_model=response_schema,
                )
            except Exception as exc:
                if attempts >= retries:
                    raise
                payload.append(
                    {
                        "role": "user",
                        "content": _structured_retry_prompt(exc),
                    }
                )
                attempts += 1



import asyncio
from collections.abc import Callable
from typing import Any, Protocol

from blackgeorge.core.job import Job
from blackgeorge.workflow.context import WorkflowContext
from blackgeorge.workflow.result import StepOutput, StepResult


class Executable(Protocol):
    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]: ...


class Step:
    def __init__(
        self,
        runner: Any,
        name: str | None = None,
        job_builder: Callable[[WorkflowContext], Job] | None = None,
    ) -> None:
        self.runner = runner
        self.name = name or getattr(runner, "name", "step")
        self.job_builder = job_builder

    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]:
        job = self.job_builder(context) if self.job_builder else context.job
        flow.emit("step.started", self.name, {})
        report, state = await flow._run_runner(self.runner, job)
        if state is not None and report.status == "paused":
            flow.emit("step.paused", self.name, {"status": report.status})
            return [StepResult(report, state)]
        flow.emit("step.completed", self.name, {"status": report.status})
        return [StepResult(report, state)]


def _should_stop(outputs: list[StepOutput]) -> bool:
    for item in outputs:
        report = item.report if isinstance(item, StepResult) else item
        if report.status in ("paused", "failed"):
            return True
    return False


class Parallel:
    def __init__(self, *steps: Executable) -> None:
        self.steps = list(steps)

    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]:
        async def run_step(step: Executable) -> list[StepOutput]:
            return await step.execute(flow, context)

        results = await asyncio.gather(*(run_step(step) for step in self.steps))
        return [report for group in results for report in group]


class Condition:
    def __init__(
        self,
        predicate: Callable[[WorkflowContext], bool],
        if_true: list[Executable],
        if_false: list[Executable] | None = None,
    ) -> None:
        self.predicate = predicate
        self.if_true = if_true
        self.if_false = if_false or []

    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]:
        steps = self.if_true if self.predicate(context) else self.if_false
        reports: list[StepOutput] = []
        for step in steps:
            step_outputs = await step.execute(flow, context)
            reports.extend(step_outputs)
            if _should_stop(step_outputs):
                return reports
        return reports


class Router:
    def __init__(
        self,
        selector: Callable[[WorkflowContext], str],
        routes: dict[str, list[Executable]],
    ) -> None:
        self.selector = selector
        self.routes = routes

    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]:
        key = self.selector(context)
        steps = self.routes.get(key, [])
        reports: list[StepOutput] = []
        for step in steps:
            step_outputs = await step.execute(flow, context)
            reports.extend(step_outputs)
            if _should_stop(step_outputs):
                return reports
        return reports


class Loop:
    def __init__(
        self,
        steps: list[Executable],
        stop: Callable[[WorkflowContext], bool],
        max_iterations: int = 10,
    ) -> None:
        self.steps = steps
        self.stop = stop
        self.max_iterations = max_iterations

    async def execute(self, flow: Any, context: WorkflowContext) -> list[StepOutput]:
        reports: list[StepOutput] = []
        iteration = 0
        while iteration < self.max_iterations:
            iteration += 1
            for step in self.steps:
                step_outputs = await step.execute(flow, context)
                reports.extend(step_outputs)
                if _should_stop(step_outputs):
                    return reports
            if self.stop(context):
                break
        return reports



from dataclasses import dataclass
from datetime import datetime
from typing import Any

from blackgeorge.core.message import Message


@dataclass(frozen=True)
class SessionRecord:
    session_id: str
    worker_name: str
    created_at: datetime
    updated_at: datetime
    metadata: dict[str, Any]


class SessionStore:
    def create_session(
        self,
        session_id: str,
        worker_name: str,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        raise NotImplementedError

    def update_session(
        self,
        session_id: str,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        raise NotImplementedError

    def get_session(self, session_id: str) -> SessionRecord | None:
        raise NotImplementedError

    def list_sessions(
        self,
        worker_name: str | None = None,
        limit: int | None = None,
    ) -> list[SessionRecord]:
        raise NotImplementedError

    def delete_session(self, session_id: str) -> None:
        raise NotImplementedError

    def add_messages(self, session_id: str, messages: list[Message]) -> None:
        raise NotImplementedError

    def get_messages(self, session_id: str) -> list[Message]:
        raise NotImplementedError



from typing import Any

import litellm
from litellm.cost_calculator import completion_cost


def calculate_cost(response: Any) -> float | None:
    try:
        return completion_cost(completion_response=response)
    except Exception:
        return None


def get_model_pricing(model: str) -> dict[str, float] | None:
    try:
        return litellm.model_cost.get(model)
    except Exception:
        return None


def get_prompt_cost(model: str, tokens: int) -> float | None:
    pricing = get_model_pricing(model)
    if not pricing:
        return None
    if "input_cost_per_million" in pricing:
        return pricing["input_cost_per_million"] * tokens / 1_000_000
    if "input_cost_per_token" in pricing:
        return pricing["input_cost_per_token"] * tokens
    return None


def get_completion_cost(model: str, tokens: int) -> float | None:
    pricing = get_model_pricing(model)
    if not pricing:
        return None
    if "output_cost_per_million" in pricing:
        return pricing["output_cost_per_million"] * tokens / 1_000_000
    if "output_cost_per_token" in pricing:
        return pricing["output_cost_per_token"] * tokens
    return None



from blackgeorge.memory.base import MemoryStore


class ExternalMemoryStore(MemoryStore):
    def write(self, key: str, value: object, scope: str) -> None:
        raise NotImplementedError

    def read(self, key: str, scope: str) -> object | None:
        raise NotImplementedError

    def search(self, query: str, scope: str) -> list[tuple[str, object]]:
        raise NotImplementedError

    def reset(self, scope: str) -> None:
        raise NotImplementedError



from collections.abc import Iterable

from blackgeorge.tools.base import Tool


class Toolbelt:
    def __init__(self, tools: Iterable[Tool] | None = None) -> None:
        self._tools: dict[str, Tool] = {}
        if tools:
            for tool in tools:
                self.add(tool)

    def add(self, tool: Tool) -> None:
        self._tools[tool.name] = tool

    def resolve(self, name: str) -> Tool | None:
        return self._tools.get(name)

    def list(self) -> list[Tool]:
        return list(self._tools.values())



import asyncio
import json
import warnings
from collections.abc import Callable, Iterable
from typing import Any, cast

from pydantic import BaseModel, TypeAdapter

from blackgeorge.adapters.base import BaseModelAdapter, ModelResponse
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.report import Report
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool, ToolResult
from blackgeorge.tools.execution import aexecute_tool
from blackgeorge.tools.registry import Toolbelt
from blackgeorge.utils import new_id
from blackgeorge.worker_context import (
    aapply_context_summary,
    is_context_limit_error,
    litellm_model_registered,
)
from blackgeorge.worker_messages import (
    chunk_content,
    chunk_reasoning_content,
    chunk_thinking_blocks,
    chunk_usage,
    emit_assistant_message,
    ensure_content,
    messages_to_payload,
    render_input,
    replace_tool_call,
    system_message,
    tool_call_with_result,
    tool_message,
    tool_schemas,
)
from blackgeorge.worker_runner_helpers import (
    EventEmitter,
    _acontext_retry,
    _build_report,
    _build_state,
    _ensure_not_running_loop,
    _execute_tool_calls_async,
    _fail_report,
    _finalize_plain_response,
    _finalize_structured_response,
    _plan_tool_calls,
    _record_usage,
    _report_error,
    _should_stream,
    _tool_event_payload,
)
from blackgeorge.worker_tools import resume_argument_key, update_arguments


def _parse_structured_stream_json(response_schema: Any, content: str) -> Any:
    if isinstance(response_schema, TypeAdapter):
        return response_schema.validate_json(content)
    if isinstance(response_schema, type) and issubclass(response_schema, BaseModel):
        return response_schema.model_validate_json(content)
    return TypeAdapter(response_schema).validate_json(content)


def _stream_value(value: Any, key: str, default: Any = None) -> Any:
    if isinstance(value, dict):
        return value.get(key, default)
    return getattr(value, key, default)


def _append_tool_error(current: str | None, message: str) -> str:
    return message if current is None else f"{current}; {message}"


def _is_stream_unsupported_error(exc: Exception) -> bool:
    if isinstance(exc, NotImplementedError):
        return True
    message = f"{type(exc).__name__}: {exc}".lower()
    mentions_stream = "stream" in message or "streaming" in message
    unsupported_markers = (
        "unsupported",
        "not support",
        "not supported",
        "not implemented",
        "cannot stream",
    )
    return mentions_stream and any(marker in message for marker in unsupported_markers)


def _chunk_tool_call_deltas(chunk: Any) -> tuple[list[Any], bool]:
    choices = _stream_value(chunk, "choices", []) or []
    if not choices:
        return [], False
    choice = choices[0]
    delta = _stream_value(choice, "delta")
    if delta is not None:
        tool_calls = _stream_value(delta, "tool_calls")
        if isinstance(tool_calls, list) and tool_calls:
            return tool_calls, False
    message = _stream_value(choice, "message")
    if message is None:
        return [], False
    tool_calls = _stream_value(message, "tool_calls", []) or []
    if isinstance(tool_calls, list):
        return tool_calls, True
    return [], False


def _streamed_tool_calls(states: list[dict[str, Any]]) -> list[ToolCall]:
    parsed: list[ToolCall] = []
    for state in states:
        error = cast(str | None, state.get("error"))
        name_raw = state.get("name")
        name = name_raw.strip() if isinstance(name_raw, str) else ""
        if not name:
            error = _append_tool_error(error, "Missing tool name")

        arguments: dict[str, Any] = {}
        arguments_obj = state.get("arguments_obj")
        if isinstance(arguments_obj, dict):
            arguments = dict(arguments_obj)
        else:
            argument_parts = state.get("arguments_parts")
            argument_text = "".join(argument_parts) if isinstance(argument_parts, list) else ""
            if argument_text:
                try:
                    parsed_arguments = json.loads(argument_text)
                    if isinstance(parsed_arguments, dict):
                        arguments = parsed_arguments
                    else:
                        error = _append_tool_error(
                            error, "Tool arguments JSON must decode to object"
                        )
                except json.JSONDecodeError as exc:
                    error = _append_tool_error(
                        error,
                        f"Invalid JSON in tool arguments: {exc}. Raw: {argument_text[:100]}",
                    )

        call_id_raw = state.get("id")
        call_id = call_id_raw if isinstance(call_id_raw, str) and call_id_raw else new_id()
        parsed.append(ToolCall(id=call_id, name=name, arguments=arguments, error=error))
    return parsed


class WorkerRunner:
    def __init__(self, name: str, toolbelt: Toolbelt, instructions: str | None) -> None:
        self.name = name
        self.toolbelt = toolbelt
        self.instructions = instructions

    def _build_messages(self, job: Job) -> list[Message]:
        messages: list[Message] = []

        if job.initial_messages:
            messages.extend(job.initial_messages)

        system_content = system_message(self.instructions, job)
        if system_content:
            if not messages or messages[0].role != "system":
                messages.insert(0, Message(role="system", content=system_content))
            else:
                messages[0] = Message(
                    role="system",
                    content=f"{messages[0].content}\n\n{system_content}",
                )

        messages.append(Message(role="user", content=render_input(job.input)))
        return messages

    def _resolve_tools(self, job: Job) -> list[Tool]:
        if job.tools_override is not None:
            resolved: list[Tool] = []
            for item in job.tools_override:
                if isinstance(item, Tool):
                    resolved.append(item)
                    continue
                if isinstance(item, str):
                    tool = self.toolbelt.resolve(item)
                    if tool is not None:
                        resolved.append(tool)
            return resolved
        return self.toolbelt.list()

    async def _astructured_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        response_schema: Any,
        retries: int,
    ) -> Any:
        payload = messages_to_payload(messages)
        try:
            return await adapter.astructured_complete(
                model=model,
                messages=payload,
                response_schema=response_schema,
                retries=retries,
            )
        except NotImplementedError:
            return await asyncio.to_thread(
                adapter.structured_complete,
                model=model,
                messages=payload,
                response_schema=response_schema,
                retries=retries,
            )

    async def _acompletion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        tools: list[Tool],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
        run_id: str | None = None,
        emit: EventEmitter | None = None,
    ) -> ModelResponse:
        if run_id and emit and hasattr(adapter, "set_callback_context"):
            adapter.set_callback_context(run_id, emit)
        try:
            try:
                response = await adapter.acomplete(
                    model=model,
                    messages=messages_to_payload(messages),
                    tools=tool_schemas(tools) if tools else None,
                    tool_choice="auto" if tools else None,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=False,
                    stream_options=stream_options,
                    thinking=thinking,
                    drop_params=drop_params,
                    extra_body=extra_body,
                )
            except NotImplementedError:
                response = await asyncio.to_thread(
                    adapter.complete,
                    model=model,
                    messages=messages_to_payload(messages),
                    tools=tool_schemas(tools) if tools else None,
                    tool_choice="auto" if tools else None,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=False,
                    stream_options=stream_options,
                    thinking=thinking,
                    drop_params=drop_params,
                    extra_body=extra_body,
                )
            if isinstance(response, ModelResponse):
                return response
            return ModelResponse(content=None, tool_calls=[], usage={}, raw=response)
        finally:
            if run_id and emit and hasattr(adapter, "clear_callback_context"):
                adapter.clear_callback_context()

    async def _astream_completion(
        self,
        *,
        adapter: BaseModelAdapter,
        model: str,
        messages: list[Message],
        tools: list[Tool],
        temperature: float | None,
        max_tokens: int | None,
        stream_options: dict[str, Any] | None,
        on_token: Callable[[str], None],
        thinking: dict[str, Any] | None = None,
        drop_params: bool | None = None,
        extra_body: dict[str, Any] | None = None,
        run_id: str | None = None,
        emit: EventEmitter | None = None,
    ) -> ModelResponse:
        if run_id and emit and hasattr(adapter, "set_callback_context"):
            adapter.set_callback_context(run_id, emit)
        try:
            try:
                stream = await adapter.acomplete(
                    model=model,
                    messages=messages_to_payload(messages),
                    tools=tool_schemas(tools) if tools else None,
                    tool_choice="auto" if tools else None,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=True,
                    stream_options=stream_options,
                    thinking=thinking,
                    drop_params=drop_params,
                    extra_body=extra_body,
                )
            except NotImplementedError:
                try:
                    stream = await asyncio.to_thread(
                        adapter.complete,
                        model=model,
                        messages=messages_to_payload(messages),
                        tools=tool_schemas(tools) if tools else None,
                        tool_choice="auto" if tools else None,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream=True,
                        stream_options=stream_options,
                        thinking=thinking,
                        drop_params=drop_params,
                        extra_body=extra_body,
                    )
                except Exception as exc:
                    if _is_stream_unsupported_error(exc):
                        return await self._acompletion(
                            adapter=adapter,
                            model=model,
                            messages=messages,
                            tools=tools,
                            temperature=temperature,
                            max_tokens=max_tokens,
                            stream_options=stream_options,
                            thinking=thinking,
                            drop_params=drop_params,
                            extra_body=extra_body,
                            run_id=run_id,
                            emit=emit,
                        )
                    raise
            except Exception as exc:
                if _is_stream_unsupported_error(exc):
                    return await self._acompletion(
                        adapter=adapter,
                        model=model,
                        messages=messages,
                        tools=tools,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        thinking=thinking,
                        drop_params=drop_params,
                        extra_body=extra_body,
                        run_id=run_id,
                        emit=emit,
                    )
                raise
            if isinstance(stream, ModelResponse):
                return stream
            is_async_stream = hasattr(stream, "__aiter__")
            is_sync_stream = hasattr(stream, "__iter__") and not isinstance(
                stream, (str, bytes, bytearray, dict)
            )
            if not is_async_stream and not is_sync_stream:
                return await self._acompletion(
                    adapter=adapter,
                    model=model,
                    messages=messages,
                    tools=tools,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream_options=stream_options,
                    thinking=thinking,
                    drop_params=drop_params,
                    extra_body=extra_body,
                    run_id=run_id,
                    emit=emit,
                )
            content_parts: list[str] = []
            reasoning_parts: list[str] = []
            thinking_blocks: list[dict[str, Any]] = []
            tool_states: list[dict[str, Any]] = []
            keyed_states: dict[tuple[str, Any], dict[str, Any]] = {}
            usage: dict[str, Any] = {}

            def process_chunk(chunk: Any) -> None:
                nonlocal usage
                token = chunk_content(chunk)
                if token:
                    content_parts.append(token)
                    on_token(token)

                tool_deltas, from_message_payload = _chunk_tool_call_deltas(chunk)
                for position, tool_delta in enumerate(tool_deltas):
                    index_value = _stream_value(tool_delta, "index")
                    call_id_value = _stream_value(tool_delta, "id")
                    stable_keys: list[tuple[str, Any]] = []
                    if isinstance(index_value, int):
                        stable_keys.append(("index", index_value))
                    if isinstance(call_id_value, str) and call_id_value:
                        stable_keys.append(("id", call_id_value))
                    fallback_key = ("position", position)
                    lookup_keys = stable_keys or [fallback_key]

                    state: dict[str, Any] | None = None
                    for key in lookup_keys:
                        existing = keyed_states.get(key)
                        if existing is not None:
                            state = existing
                            break
                    if state is None:
                        state = {
                            "id": None,
                            "name": "",
                            "arguments_parts": [],
                            "arguments_obj": None,
                            "error": None,
                        }
                        tool_states.append(state)
                    store_keys = stable_keys or [fallback_key]
                    for key in store_keys:
                        keyed_states[key] = state

                    if isinstance(call_id_value, str) and call_id_value:
                        state["id"] = call_id_value

                    function = _stream_value(tool_delta, "function")
                    name_value = _stream_value(function, "name")
                    if isinstance(name_value, str):
                        name = name_value.strip()
                        if name:
                            existing_name = cast(str, state["name"])
                            if not existing_name or name.startswith(existing_name):
                                state["name"] = name
                            elif not existing_name.startswith(name):
                                state["name"] = f"{existing_name}{name}"

                    arguments_value = _stream_value(function, "arguments")
                    if isinstance(arguments_value, str):
                        argument_parts = cast(list[str], state["arguments_parts"])
                        if from_message_payload:
                            previous_arguments = "".join(argument_parts)
                            if previous_arguments != arguments_value:
                                argument_parts.clear()
                                argument_parts.append(arguments_value)
                                if arguments_value:
                                    on_token(arguments_value)
                        else:
                            argument_parts.append(arguments_value)
                            if arguments_value:
                                on_token(arguments_value)
                    elif isinstance(arguments_value, dict):
                        arguments_obj = state.get("arguments_obj")
                        if isinstance(arguments_obj, dict):
                            merged_arguments = dict(arguments_obj)
                        else:
                            merged_arguments = {}
                        merged_arguments.update(arguments_value)
                        state["arguments_obj"] = merged_arguments
                        serialized = json.dumps(arguments_value, ensure_ascii=True)
                        if serialized:
                            on_token(serialized)
                    elif arguments_value is not None:
                        type_name = type(arguments_value).__name__
                        state["error"] = _append_tool_error(
                            cast(str | None, state.get("error")),
                            f"Unsupported tool arguments type: {type_name}",
                        )

                reasoning = chunk_reasoning_content(chunk)
                if reasoning:
                    reasoning_parts.append(reasoning)
                blocks = chunk_thinking_blocks(chunk)
                if blocks:
                    thinking_blocks.extend(blocks)
                usage_chunk = chunk_usage(chunk)
                if usage_chunk:
                    usage = usage_chunk

            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Pydantic serializer warnings")
                try:
                    if is_async_stream:
                        async for chunk in stream:
                            process_chunk(chunk)
                    else:
                        for chunk in cast(Iterable[Any], stream):
                            process_chunk(chunk)
                finally:
                    if hasattr(stream, "aclose"):
                        await stream.aclose()
                    elif hasattr(stream, "close"):
                        stream.close()
            return ModelResponse(
                content="".join(content_parts),
                reasoning_content="".join(reasoning_parts) or None,
                thinking_blocks=thinking_blocks or None,
                tool_calls=_streamed_tool_calls(tool_states),
                usage=usage,
                raw=stream,
            )
        finally:
            if run_id and emit and hasattr(adapter, "clear_callback_context"):
                adapter.clear_callback_context()

    async def _arun_loop(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        messages: list[Message],
        tool_calls: list[ToolCall],
        metrics: dict[str, Any],
        errors: list[str],
        iteration: int,
        model_name: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        tools = self._resolve_tools(job)
        allowed_tools = {tool.name: tool for tool in tools}
        response_schema = job.response_schema
        structured_stream_mode = job.structured_stream_mode or "off"
        context_summaries = 0
        model_registered = litellm_model_registered(model_name)

        while iteration < max_iterations:
            iteration += 1
            if _should_stream(stream, tools, response_schema):
                try:
                    response = await self._astream_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        tools=tools,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        on_token=lambda token: emit("stream.token", self.name, {"token": token}),
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                        run_id=run_id,
                        emit=emit,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    decision = await _acontext_retry(
                        run_id=run_id,
                        worker_name=self.name,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        model_registered=model_registered,
                        respect_context_window=respect_context_window,
                        context_summaries=context_summaries,
                        apply_summary=lambda: aapply_context_summary(
                            adapter=adapter,
                            model_name=model_name,
                            messages=messages,
                            temperature=temperature,
                            metrics=metrics,
                            emit=emit,
                            worker_name=self.name,
                            model_registered=model_registered,
                        ),
                    )
                    if decision.report is not None:
                        return decision.report, None
                    context_summaries += 1
                    continue
            elif (
                stream
                and response_schema is not None
                and not tools
                and structured_stream_mode == "preview"
            ):
                try:
                    streamed = await self._astream_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        tools=[],
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        on_token=lambda token: emit("stream.token", self.name, {"token": token}),
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                        run_id=run_id,
                        emit=emit,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    decision = await _acontext_retry(
                        run_id=run_id,
                        worker_name=self.name,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        model_registered=model_registered,
                        respect_context_window=respect_context_window,
                        context_summaries=context_summaries,
                        apply_summary=lambda: aapply_context_summary(
                            adapter=adapter,
                            model_name=model_name,
                            messages=messages,
                            temperature=temperature,
                            metrics=metrics,
                            emit=emit,
                            worker_name=self.name,
                            model_registered=model_registered,
                        ),
                    )
                    if decision.report is not None:
                        return decision.report, None
                    context_summaries += 1
                    continue

                metrics["usage"] = streamed.usage
                streamed_content = streamed.content or ""
                try:
                    data = _parse_structured_stream_json(response_schema, streamed_content)
                except Exception:
                    try:
                        data = await self._astructured_completion(
                            adapter=adapter,
                            model=model_name,
                            messages=messages,
                            response_schema=response_schema,
                            retries=structured_output_retries,
                        )
                    except Exception as exc:
                        if is_context_limit_error(exc):
                            decision = await _acontext_retry(
                                run_id=run_id,
                                worker_name=self.name,
                                messages=messages,
                                tool_calls=tool_calls,
                                metrics=metrics,
                                events=events,
                                errors=errors,
                                emit=emit,
                                model_registered=model_registered,
                                respect_context_window=respect_context_window,
                                context_summaries=context_summaries,
                                apply_summary=lambda: aapply_context_summary(
                                    adapter=adapter,
                                    model_name=model_name,
                                    messages=messages,
                                    temperature=temperature,
                                    metrics=metrics,
                                    emit=emit,
                                    worker_name=self.name,
                                    model_registered=model_registered,
                                ),
                            )
                            if decision.report is not None:
                                return decision.report, None
                            context_summaries += 1
                            continue
                        return (
                            _fail_report(
                                run_id=run_id,
                                worker_name=self.name,
                                message=str(exc),
                                messages=messages,
                                tool_calls=tool_calls,
                                metrics=metrics,
                                events=events,
                                errors=errors,
                                emit=emit,
                            ),
                            None,
                        )

                return (
                    _finalize_structured_response(
                        run_id=run_id,
                        data=data,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        worker_name=self.name,
                    ),
                    None,
                )
            elif response_schema is not None and not tools:
                try:
                    data = await self._astructured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    if is_context_limit_error(exc):
                        decision = await _acontext_retry(
                            run_id=run_id,
                            worker_name=self.name,
                            messages=messages,
                            tool_calls=tool_calls,
                            metrics=metrics,
                            events=events,
                            errors=errors,
                            emit=emit,
                            model_registered=model_registered,
                            respect_context_window=respect_context_window,
                            context_summaries=context_summaries,
                            apply_summary=lambda: aapply_context_summary(
                                adapter=adapter,
                                model_name=model_name,
                                messages=messages,
                                temperature=temperature,
                                metrics=metrics,
                                emit=emit,
                                worker_name=self.name,
                                model_registered=model_registered,
                            ),
                        )
                        if decision.report is not None:
                            return decision.report, None
                        context_summaries += 1
                        continue
                    return (
                        _fail_report(
                            run_id=run_id,
                            worker_name=self.name,
                            message=str(exc),
                            messages=messages,
                            tool_calls=tool_calls,
                            metrics=metrics,
                            events=events,
                            errors=errors,
                            emit=emit,
                        ),
                        None,
                    )
                return (
                    _finalize_structured_response(
                        run_id=run_id,
                        data=data,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        worker_name=self.name,
                    ),
                    None,
                )
            else:
                try:
                    response = await self._acompletion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        tools=tools,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stream_options=stream_options,
                        thinking=job.thinking,
                        drop_params=job.drop_params,
                        extra_body=job.extra_body,
                        run_id=run_id,
                        emit=emit,
                    )
                except Exception as exc:
                    if not is_context_limit_error(exc):
                        raise
                    decision = await _acontext_retry(
                        run_id=run_id,
                        worker_name=self.name,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        model_registered=model_registered,
                        respect_context_window=respect_context_window,
                        context_summaries=context_summaries,
                        apply_summary=lambda: aapply_context_summary(
                            adapter=adapter,
                            model_name=model_name,
                            messages=messages,
                            temperature=temperature,
                            metrics=metrics,
                            emit=emit,
                            worker_name=self.name,
                            model_registered=model_registered,
                        ),
                    )
                    if decision.report is not None:
                        return decision.report, None
                    context_summaries += 1
                    continue

            _record_usage(metrics, response)

            if response.tool_calls:
                assistant_message = Message(
                    role="assistant",
                    content=ensure_content(response.content),
                    reasoning_content=response.reasoning_content,
                    thinking_blocks=response.thinking_blocks,
                    tool_calls=response.tool_calls,
                )
                messages.append(assistant_message)
                emit_assistant_message(emit, self.name, assistant_message)
                plan = _plan_tool_calls(
                    response=response,
                    allowed_tools=allowed_tools,
                    tool_calls=tool_calls,
                    max_tool_calls=max_tool_calls,
                )

                await _execute_tool_calls_async(
                    plan.ordered_calls,
                    plan.executable_calls,
                    plan.immediate_results,
                    messages,
                    tool_calls,
                    emit,
                )

                if plan.max_tool_calls_exceeded:
                    return (
                        _fail_report(
                            run_id=run_id,
                            worker_name=self.name,
                            message="Max tool calls exceeded",
                            messages=messages,
                            tool_calls=tool_calls,
                            metrics=metrics,
                            events=events,
                            errors=errors,
                            emit=emit,
                        ),
                        None,
                    )

                if plan.pending is not None:
                    emit(
                        f"tool.{plan.pending.type}_requested",
                        plan.pending.tool_call.name,
                        {"tool_call_id": plan.pending.tool_call.id},
                    )
                    emit(
                        "worker.paused",
                        self.name,
                        {"pending_action_type": plan.pending.type},
                    )
                    report = _build_report(
                        run_id,
                        "paused",
                        None,
                        None,
                        None,
                        messages,
                        tool_calls,
                        metrics,
                        events,
                        plan.pending,
                        errors,
                    )
                    state = _build_state(
                        run_id,
                        "paused",
                        self.name,
                        job,
                        messages,
                        tool_calls,
                        plan.pending,
                        metrics,
                        iteration,
                    )
                    return report, state
                continue

            if response_schema is not None:
                try:
                    data = await self._astructured_completion(
                        adapter=adapter,
                        model=model_name,
                        messages=messages,
                        response_schema=response_schema,
                        retries=structured_output_retries,
                    )
                except Exception as exc:
                    if is_context_limit_error(exc):
                        decision = await _acontext_retry(
                            run_id=run_id,
                            worker_name=self.name,
                            messages=messages,
                            tool_calls=tool_calls,
                            metrics=metrics,
                            events=events,
                            errors=errors,
                            emit=emit,
                            model_registered=model_registered,
                            respect_context_window=respect_context_window,
                            context_summaries=context_summaries,
                            apply_summary=lambda: aapply_context_summary(
                                adapter=adapter,
                                model_name=model_name,
                                messages=messages,
                                temperature=temperature,
                                metrics=metrics,
                                emit=emit,
                                worker_name=self.name,
                                model_registered=model_registered,
                            ),
                        )
                        if decision.report is not None:
                            return decision.report, None
                        context_summaries += 1
                        continue
                    return (
                        _fail_report(
                            run_id=run_id,
                            worker_name=self.name,
                            message=str(exc),
                            messages=messages,
                            tool_calls=tool_calls,
                            metrics=metrics,
                            events=events,
                            errors=errors,
                            emit=emit,
                        ),
                        None,
                    )
                return (
                    _finalize_structured_response(
                        run_id=run_id,
                        data=data,
                        messages=messages,
                        tool_calls=tool_calls,
                        metrics=metrics,
                        events=events,
                        errors=errors,
                        emit=emit,
                        worker_name=self.name,
                    ),
                    None,
                )

            return (
                _finalize_plain_response(
                    run_id=run_id,
                    response=response,
                    messages=messages,
                    tool_calls=tool_calls,
                    metrics=metrics,
                    events=events,
                    errors=errors,
                    emit=emit,
                    worker_name=self.name,
                ),
                None,
            )

        return _fail_report(
            run_id=run_id,
            worker_name=self.name,
            message="Max iterations exceeded",
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            events=events,
            errors=errors,
            emit=emit,
        ), None

    def run(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        _ensure_not_running_loop("run", "arun")
        return asyncio.run(
            self.arun(
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                model_name=model_name,
                respect_context_window=respect_context_window,
            )
        )

    async def arun(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        messages = self._build_messages(job)
        tool_calls: list[ToolCall] = []
        metrics: dict[str, Any] = {}
        errors: list[str] = []
        iteration = 0

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return await self._arun_loop(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    def resume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        _ensure_not_running_loop("resume", "aresume")
        return asyncio.run(
            self.aresume(
                adapter=adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                model_name=model_name,
                respect_context_window=respect_context_window,
            )
        )

    async def aresume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        pending = state.pending_action
        if pending is None:
            report = _build_report(
                state.run_id,
                "failed",
                None,
                None,
                None,
                state.messages,
                state.tool_calls,
                state.metrics,
                events,
                None,
                ["No pending action"],
            )
            return report, None

        messages = list(state.messages)
        tool_calls = list(state.tool_calls)
        iteration = state.iteration
        metrics = dict(state.metrics)
        errors: list[str] = []

        tool = self.toolbelt.resolve(pending.tool_call.name)
        if tool is None:
            result = ToolResult(error=f"Tool not found: {pending.tool_call.name}")
            emit(
                "tool.failed",
                pending.tool_call.name,
                {"tool_call_id": pending.tool_call.id, "error": result.error},
            )
            messages.append(tool_message(result, pending.tool_call))
            replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
        else:
            if pending.type == "confirmation" and not decision_or_input:
                result = ToolResult(error="Tool execution declined")
                emit(
                    "tool.failed",
                    pending.tool_call.name,
                    {"tool_call_id": pending.tool_call.id, "error": result.error},
                )
                tool_result_message = tool_message(result, pending.tool_call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(pending.tool_call, result))
            else:
                call = pending.tool_call
                if pending.type == "user_input":
                    key = resume_argument_key(pending)
                    call = update_arguments(call, key, decision_or_input)
                emit("tool.started", tool.name, {"tool_call_id": call.id})
                result = await aexecute_tool(tool, call)
                if result.error:
                    emit("tool.failed", tool.name, {"tool_call_id": call.id, "error": result.error})
                else:
                    emit("tool.completed", tool.name, _tool_event_payload(call, result))
                tool_result_message = tool_message(result, call)
                messages.append(tool_result_message)
                replace_tool_call(tool_calls, tool_call_with_result(call, result))

        if not model_name:
            errors.append("Worker model not set")
            emit("worker.failed", self.name, {"error": errors[-1]})
            report = _report_error(state.run_id, messages, errors, events)
            return report, None

        emit("worker.started", self.name, {})
        return await self._arun_loop(
            adapter=adapter,
            job=state.job,
            run_id=state.run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            messages=messages,
            tool_calls=tool_calls,
            metrics=metrics,
            errors=errors,
            iteration=iteration,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )



import json
from collections.abc import Callable
from dataclasses import asdict, is_dataclass
from typing import Any, cast

from pydantic import BaseModel

from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.tools.base import Tool, ToolResult


def _json_payload(value: Any) -> Any:
    if isinstance(value, BaseModel):
        return value.model_dump(mode="json", warnings=False)
    if is_dataclass(value) and not isinstance(value, type):
        return asdict(value)
    if isinstance(value, dict):
        return {key: _json_payload(item) for key, item in value.items()}
    if isinstance(value, list):
        return [_json_payload(item) for item in value]
    if isinstance(value, tuple):
        return [_json_payload(item) for item in value]
    return value


def _is_multimodal_content_blocks(value: Any) -> bool:
    if not isinstance(value, list):
        return False
    if len(value) == 0:
        return False
    known_type_fields: dict[str, tuple[str, ...]] = {
        "text": ("text",),
        "image_url": ("image_url",),
        "video_url": ("video_url",),
        "file": ("file",),
        "audio_url": ("audio_url",),
        "input_text": ("text",),
        "input_image": ("image_url", "file_id"),
        "input_audio": ("input_audio",),
    }
    for item in value:
        if not isinstance(item, dict):
            return False
        block_type = item.get("type")
        if not isinstance(block_type, str):
            return False
        required_fields = known_type_fields.get(block_type)
        if required_fields is None:
            return False
        if not any(field in item for field in required_fields):
            return False
    return True


def render_input(value: Any) -> str | list[dict[str, Any]]:
    if isinstance(value, str):
        return value
    if _is_multimodal_content_blocks(value):
        payload = _json_payload(value)
        if isinstance(payload, list):
            return cast(list[dict[str, Any]], payload)
    return json.dumps(_json_payload(value), ensure_ascii=True, default=str)


def tool_call_payload(tool_call: ToolCall) -> dict[str, Any]:
    return {
        "id": tool_call.id,
        "type": "function",
        "function": {
            "name": tool_call.name,
            "arguments": json.dumps(
                _json_payload(tool_call.arguments),
                ensure_ascii=True,
                default=str,
            ),
        },
    }


def messages_to_payload(messages: list[Message]) -> list[dict[str, Any]]:
    payload: list[dict[str, Any]] = []
    for message in messages:
        item: dict[str, Any] = {"role": message.role, "content": message.content}
        if message.role == "assistant":
            if message.reasoning_content is not None:
                item["reasoning_content"] = message.reasoning_content
            if message.thinking_blocks is not None:
                item["thinking_blocks"] = message.thinking_blocks
            if message.tool_calls:
                item["tool_calls"] = [tool_call_payload(call) for call in message.tool_calls]
        if message.role == "tool" and message.tool_call_id:
            item["tool_call_id"] = message.tool_call_id
        payload.append(item)
    return payload


def tool_schemas(tools: list[Tool]) -> list[dict[str, Any]]:
    schemas: list[dict[str, Any]] = []
    for tool in tools:
        schemas.append(
            {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": tool.schema,
                },
            }
        )
    return schemas


def system_message(instructions: str | None, job: Job) -> str | None:
    parts: list[str] = []
    if instructions:
        parts.append(instructions)
    if job.expected_output:
        parts.append(f"Expected output: {job.expected_output}")
    if job.constraints:
        parts.append(
            "Constraints: "
            f"{json.dumps(_json_payload(job.constraints), ensure_ascii=True, default=str)}"
        )
    if not parts:
        return None
    return "\n\n".join(parts)


def chunk_content(chunk: Any) -> str | None:
    if isinstance(chunk, dict):
        choices = chunk.get("choices") or []
        if not choices:
            return None
        delta = choices[0].get("delta") or {}
        return delta.get("content")
    choices = getattr(chunk, "choices", [])
    if not choices:
        return None
    delta = getattr(choices[0], "delta", None)
    if delta is None:
        return None
    return getattr(delta, "content", None)


def chunk_usage(chunk: Any) -> dict[str, Any] | None:
    if isinstance(chunk, dict):
        return chunk.get("usage")
    return getattr(chunk, "usage", None)


def chunk_reasoning_content(chunk: Any) -> str | None:
    if isinstance(chunk, dict):
        choices = chunk.get("choices") or []
        if not choices:
            return None
        delta = choices[0].get("delta") or {}
        return delta.get("reasoning_content")
    choices = getattr(chunk, "choices", [])
    if not choices:
        return None
    delta = getattr(choices[0], "delta", None)
    if delta is None:
        return None
    return getattr(delta, "reasoning_content", None)


def chunk_thinking_blocks(chunk: Any) -> list[dict[str, Any]] | None:
    if isinstance(chunk, dict):
        choices = chunk.get("choices") or []
        if not choices:
            return None
        delta = choices[0].get("delta") or {}
        blocks = delta.get("thinking_blocks")
    else:
        choices = getattr(chunk, "choices", [])
        if not choices:
            return None
        delta = getattr(choices[0], "delta", None)
        if delta is None:
            return None
        blocks = getattr(delta, "thinking_blocks", None)
    if isinstance(blocks, dict):
        return [dict(blocks)]
    if isinstance(blocks, list):
        items = [dict(item) for item in blocks if isinstance(item, dict)]
        return items or None
    return None


def structured_content(value: Any) -> str:
    if isinstance(value, BaseModel):
        return value.model_dump_json(warnings=False)
    return json.dumps(_json_payload(value), ensure_ascii=True, default=str)


def tool_message(result: ToolResult | dict[str, Any], tool_call: ToolCall) -> Message:
    if isinstance(result, ToolResult):
        if result.content is not None:
            content = result.content
        elif result.data is not None:
            try:
                content = json.dumps(result.data, ensure_ascii=True, default=str)
            except TypeError:
                content = str(result.data)
        elif result.error is not None:
            content = result.error
        else:
            content = ""
    else:
        try:
            content = json.dumps(result, ensure_ascii=True, default=str)
        except TypeError:
            content = str(result)
    return Message(role="tool", content=content, tool_call_id=tool_call.id)


def tool_call_with_result(call: ToolCall, result: ToolResult | dict[str, Any]) -> ToolCall:
    error = result.error if isinstance(result, ToolResult) else None
    return call.model_copy(update={"result": result, "error": error})


def replace_tool_call(tool_calls: list[ToolCall], updated: ToolCall) -> None:
    for index, call in enumerate(tool_calls):
        if call.id == updated.id:
            tool_calls[index] = updated
            return
    raise ValueError(f"Tool call {updated.id} not found")


def tool_call_summaries(tool_calls: list[ToolCall]) -> list[dict[str, Any]]:
    summaries: list[dict[str, Any]] = []
    for call in tool_calls:
        summaries.append(
            {
                "id": call.id,
                "name": call.name,
                "arguments": call.arguments,
            }
        )
    return summaries


def emit_assistant_message(
    emit: Callable[[str, str, dict[str, Any]], None],
    worker_name: str,
    message: Message,
) -> None:
    payload: dict[str, Any] = {"content": message.content}
    if message.tool_calls:
        payload["tool_calls"] = tool_call_summaries(message.tool_calls)
    emit("assistant.message", worker_name, payload)


def ensure_content(value: str | None) -> str:
    return value or ""



from dataclasses import asdict, is_dataclass
from typing import Any, cast

from pydantic import BaseModel, ConfigDict, field_serializer


class ToolCall(BaseModel):
    id: str
    name: str
    arguments: dict[str, Any]
    result: Any | None = None
    error: str | None = None

    model_config = ConfigDict(frozen=True)

    @field_serializer("result")
    def _serialize_result(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, BaseModel):
            return value.model_dump(mode="json", warnings=False)
        if is_dataclass(value) and not isinstance(value, type):
            return asdict(cast(Any, value))
        return value



from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import PendingActionType


class PendingAction(BaseModel):
    action_id: str
    type: PendingActionType
    tool_call: ToolCall
    prompt: str
    options: list[str] = Field(default_factory=list)
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)



from blackgeorge.collaboration.blackboard import Blackboard
from blackgeorge.collaboration.channel import BroadcastMode, Channel
from blackgeorge.tools import Tool, tool

type JsonValue = str | int | float | bool | None | list[object] | dict[str, object]


def channel_send_tool(channel: Channel, sender: str, name: str = "channel_send") -> Tool:
    @tool(name=name)
    def channel_send(
        recipient: str,
        content: JsonValue,
        metadata: dict[str, JsonValue] | None = None,
    ) -> str:
        message = channel.send(sender, recipient, content, metadata)
        return message.id

    return channel_send


def channel_broadcast_tool(channel: Channel, sender: str, name: str = "channel_broadcast") -> Tool:
    @tool(name=name)
    def channel_broadcast(
        content: JsonValue,
        metadata: dict[str, JsonValue] | None = None,
    ) -> str:
        message = channel.broadcast(sender, content, metadata)
        return message.id

    return channel_broadcast


def channel_receive_tool(
    channel: Channel,
    recipient: str,
    name: str = "channel_receive",
) -> Tool:
    @tool(name=name)
    def channel_receive(
        broadcast_mode: BroadcastMode = "one_shot",
        clear: bool = True,
    ) -> list[dict[str, JsonValue]]:
        messages = channel.receive(recipient, clear=clear, broadcast_mode=broadcast_mode)
        output: list[dict[str, JsonValue]] = []
        for msg in messages:
            output.append(
                {
                    "id": msg.id,
                    "sender": msg.sender,
                    "recipient": msg.recipient,
                    "content": msg.content,
                    "timestamp": msg.timestamp.isoformat(),
                    "metadata": msg.metadata,
                }
            )
        return output

    return channel_receive


def blackboard_write_tool(
    blackboard: Blackboard,
    author: str,
    name: str = "blackboard_write",
) -> Tool:
    @tool(name=name)
    def blackboard_write(key: str, value: JsonValue) -> str:
        blackboard.write(key, value, author)
        return key

    return blackboard_write


def blackboard_read_tool(blackboard: Blackboard, name: str = "blackboard_read") -> Tool:
    @tool(name=name)
    def blackboard_read(key: str) -> JsonValue | None:
        return blackboard.read(key)

    return blackboard_read



import asyncio
from collections.abc import Awaitable, Callable
from typing import Any

from pydantic import BaseModel

from blackgeorge.collaboration.blackboard import Blackboard
from blackgeorge.collaboration.channel import Channel
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.report import Report
from blackgeorge.core.types import WorkforceMode
from blackgeorge.store.state import RunState
from blackgeorge.worker import Worker
from blackgeorge.workforce_helpers import (
    aggregate_reports as _aggregate_reports,
)
from blackgeorge.workforce_helpers import (
    build_workforce_state as _build_workforce_state,
)
from blackgeorge.workforce_helpers import (
    default_reducer as _default_reducer,
)
from blackgeorge.workforce_helpers import (
    find_worker as _find_worker,
)
from blackgeorge.workforce_helpers import (
    root_job as _root_job,
)
from blackgeorge.workforce_helpers import (
    select_worker_name as _select_worker_name,
)

Reducer = Callable[[list[Report]], Report]


def _ensure_not_running_loop(action: str, async_action: str) -> None:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return
    raise RuntimeError(
        f"{action} cannot be called from a running event loop. Use {async_action} instead."
    )


class WorkerDecision(BaseModel):
    worker: str
    reason: str | None = None


class Workforce:
    def __init__(
        self,
        workers: list[Worker],
        mode: WorkforceMode = "managed",
        name: str | None = None,
        manager: Worker | None = None,
        reducer: Reducer | None = None,
        channel: Channel | None = None,
        blackboard: Blackboard | None = None,
    ) -> None:
        if not workers:
            raise ValueError("Workforce requires at least one worker")
        self.workers = workers
        self.mode = mode
        self.name = name or "workforce"
        self.manager = manager
        self.reducer = reducer
        self.channel = channel or Channel()
        self.blackboard = blackboard or Blackboard()

    def _run_worker(
        self,
        *,
        worker: Worker,
        adapter: Any,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        model_name = worker.model or default_model
        return worker.run(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    def _resume_worker(
        self,
        *,
        worker: Worker,
        adapter: Any,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        model_name = worker.model or default_model
        return worker.resume(
            adapter=adapter,
            state=state,
            decision_or_input=decision_or_input,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def _arun_worker(
        self,
        *,
        worker: Worker,
        adapter: Any,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        model_name = worker.model or default_model
        return await worker.arun(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def _aresume_worker(
        self,
        *,
        worker: Worker,
        adapter: Any,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool,
    ) -> tuple[Report, RunState | None]:
        model_name = worker.model or default_model
        return await worker.aresume(
            adapter=adapter,
            state=state,
            decision_or_input=decision_or_input,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    def _can_parallelize_collaborate(self, job: Job) -> bool:
        if job.tools_override is not None:
            return len(job.tools_override) == 0
        return all(not worker.tools() for worker in self.workers)

    async def _arun_collaborate_parallel(
        self,
        *,
        adapter: Any,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool,
        drain_async_handlers: Callable[[], Awaitable[None]] | None = None,
    ) -> tuple[Report, RunState | None]:
        try:
            tasks: list[asyncio.Task[tuple[Report, RunState | None]]] = []
            for worker in self.workers:
                tasks.append(
                    asyncio.create_task(
                        self._arun_worker(
                            worker=worker,
                            adapter=adapter,
                            job=job,
                            run_id=run_id,
                            events=events,
                            emit=emit,
                            temperature=temperature,
                            max_tokens=max_tokens,
                            stream=stream,
                            stream_options=stream_options,
                            structured_output_retries=structured_output_retries,
                            max_iterations=max_iterations,
                            max_tool_calls=max_tool_calls,
                            default_model=default_model,
                            respect_context_window=respect_context_window,
                        )
                    )
                )
            results = await asyncio.gather(*tasks)
            reports: list[tuple[Worker, Report]] = []
            any_failed = False
            for worker, (report, worker_state) in zip(self.workers, results, strict=False):
                if worker_state is not None:
                    state = _build_workforce_state(
                        run_id,
                        "paused",
                        self.name,
                        job,
                        worker_state,
                        "collaborate",
                        payload={
                            "root_job": job.model_dump(mode="json"),
                            "completed_reports": [
                                rep.model_dump(mode="json") for _, rep in reports
                            ],
                            "pending_worker_index": len(reports),
                        },
                    )
                    return report, state
                reports.append((worker, report))
                if report.status == "failed":
                    any_failed = True
            if any_failed:
                return _aggregate_reports(reports, run_id, events, "failed"), None
            if self.reducer:
                return self.reducer([report for _, report in reports]), None
            return _default_reducer(reports, run_id, events), None
        finally:
            if drain_async_handlers is not None:
                await drain_async_handlers()

    def run(
        self,
        *,
        adapter: Any,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool = True,
        drain_async_handlers: Callable[[], Awaitable[None]] | None = None,
    ) -> tuple[Report, RunState | None]:
        emit("workforce.started", self.name, {})

        if self.mode == "managed":
            manager = self.manager or self.workers[0]
            manager_job = Job(
                input={
                    "task": job.input,
                    "workers": [worker.name for worker in self.workers],
                },
                response_schema=WorkerDecision,
                tools_override=[],
            )
            manager_report, manager_state = self._run_worker(
                worker=manager,
                adapter=adapter,
                job=manager_job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if manager_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    manager_state,
                    "manager",
                    payload={"root_job": job.model_dump(mode="json")},
                )
                emit("workforce.completed", self.name, {})
                return manager_report, state
            if manager_report.status == "failed":
                emit("workforce.completed", self.name, {})
                return manager_report, None

            selected = _select_worker_name(manager_report, self.workers)
            worker = _find_worker(self.workers, selected)
            report, worker_state = self._run_worker(
                worker=worker,
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if worker_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    worker_state,
                    "worker",
                    payload={
                        "root_job": job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                emit("workforce.completed", self.name, {})
                return report, state
            if report.status == "failed":
                emit("workforce.completed", self.name, {})
                return report, None
            emit("workforce.completed", self.name, {})
            return report, None

        if self._can_parallelize_collaborate(job):
            _ensure_not_running_loop("run", "arun")
            report, parallel_state = asyncio.run(
                self._arun_collaborate_parallel(
                    adapter=adapter,
                    job=job,
                    run_id=run_id,
                    events=events,
                    emit=emit,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=stream,
                    stream_options=stream_options,
                    structured_output_retries=structured_output_retries,
                    max_iterations=max_iterations,
                    max_tool_calls=max_tool_calls,
                    default_model=default_model,
                    respect_context_window=respect_context_window,
                    drain_async_handlers=drain_async_handlers,
                )
            )
            emit("workforce.completed", self.name, {})
            return report, parallel_state

        reports: list[tuple[Worker, Report]] = []
        for worker in self.workers:
            report, worker_state = self._run_worker(
                worker=worker,
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if worker_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    worker_state,
                    "collaborate",
                    payload={
                        "root_job": job.model_dump(mode="json"),
                        "completed_reports": [rep.model_dump(mode="json") for _, rep in reports],
                        "pending_worker_index": len(reports),
                    },
                )
                emit("workforce.completed", self.name, {})
                return report, state
            if report.status == "failed":
                emit("workforce.completed", self.name, {})
                return _aggregate_reports(
                    reports + [(worker, report)],
                    run_id,
                    events,
                    "failed",
                ), None
            reports.append((worker, report))

        emit("workforce.completed", self.name, {})
        if self.reducer:
            return self.reducer([report for _, report in reports]), None
        return _default_reducer(reports, run_id, events), None

    async def arun(
        self,
        *,
        adapter: Any,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        emit("workforce.started", self.name, {})

        if self.mode == "managed":
            manager = self.manager or self.workers[0]
            manager_job = Job(
                input={
                    "task": job.input,
                    "workers": [worker.name for worker in self.workers],
                },
                response_schema=WorkerDecision,
                tools_override=[],
            )
            manager_report, manager_state = await self._arun_worker(
                worker=manager,
                adapter=adapter,
                job=manager_job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if manager_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    manager_state,
                    "manager",
                    payload={"root_job": job.model_dump(mode="json")},
                )
                emit("workforce.completed", self.name, {})
                return manager_report, state
            if manager_report.status == "failed":
                emit("workforce.completed", self.name, {})
                return manager_report, None

            selected = _select_worker_name(manager_report, self.workers)
            worker = _find_worker(self.workers, selected)
            report, worker_state = await self._arun_worker(
                worker=worker,
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if worker_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    worker_state,
                    "worker",
                    payload={
                        "root_job": job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                emit("workforce.completed", self.name, {})
                return report, state
            if report.status == "failed":
                emit("workforce.completed", self.name, {})
                return report, None
            emit("workforce.completed", self.name, {})
            return report, None

        if self._can_parallelize_collaborate(job):
            report, parallel_state = await self._arun_collaborate_parallel(
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            emit("workforce.completed", self.name, {})
            return report, parallel_state

        reports: list[tuple[Worker, Report]] = []
        for worker in self.workers:
            report, worker_state = await self._arun_worker(
                worker=worker,
                adapter=adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if worker_state is not None:
                state = _build_workforce_state(
                    run_id,
                    "paused",
                    self.name,
                    job,
                    worker_state,
                    "collaborate",
                    payload={
                        "root_job": job.model_dump(mode="json"),
                        "completed_reports": [rep.model_dump(mode="json") for _, rep in reports],
                        "pending_worker_index": len(reports),
                    },
                )
                emit("workforce.completed", self.name, {})
                return report, state
            if report.status == "failed":
                emit("workforce.completed", self.name, {})
                return _aggregate_reports(
                    reports + [(worker, report)],
                    run_id,
                    events,
                    "failed",
                ), None
            reports.append((worker, report))

        emit("workforce.completed", self.name, {})
        if self.reducer:
            return self.reducer([report for _, report in reports]), None
        return _default_reducer(reports, run_id, events), None

    def resume(
        self,
        *,
        adapter: Any,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        payload = state.payload
        stage = payload.get("stage")
        worker_state_payload = payload.get("worker_state")
        if worker_state_payload is None:
            report = Report(
                run_id=state.run_id,
                status="failed",
                content=None,
                data=None,
                messages=state.messages,
                tool_calls=state.tool_calls,
                metrics=state.metrics,
                events=events,
                pending_action=None,
                errors=["Missing worker state"],
            )
            return report, None
        stored_worker_state = RunState.model_validate(worker_state_payload)

        if stage == "manager":
            manager = self.manager or self.workers[0]
            manager_report, manager_state = self._resume_worker(
                worker=manager,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if manager_state is not None:
                root_job = _root_job(payload, state.job)
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    manager_state,
                    "manager",
                    payload={"root_job": root_job.model_dump(mode="json")},
                )
                return manager_report, state
            if manager_report.status == "failed":
                return manager_report, None

            root_job = _root_job(payload, state.job)
            selected = _select_worker_name(manager_report, self.workers)
            worker = _find_worker(self.workers, selected)
            report, next_state = self._run_worker(
                worker=worker,
                adapter=adapter,
                job=root_job,
                run_id=state.run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "worker",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                return report, state
            if report.status == "failed":
                return report, None
            return report, None

        if stage == "worker":
            root_job = _root_job(payload, state.job)
            worker_name = payload.get("selected_worker")
            worker = _find_worker(self.workers, worker_name)
            report, next_state = self._resume_worker(
                worker=worker,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "worker",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                return report, state
            if report.status == "failed":
                return report, None
            return report, None

        if stage == "collaborate":
            root_job = _root_job(payload, state.job)
            completed_reports_payload = payload.get("completed_reports", [])
            completed_reports = [Report.model_validate(rep) for rep in completed_reports_payload]
            pending_index = payload.get("pending_worker_index", 0)
            if not isinstance(pending_index, int) or pending_index < 0:
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=events,
                    pending_action=None,
                    errors=["Invalid pending worker index"],
                )
                return report, None
            if pending_index >= len(self.workers):
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=events,
                    pending_action=None,
                    errors=["Invalid pending worker index"],
                )
                return report, None
            pending_worker = self.workers[pending_index]
            report, next_state = self._resume_worker(
                worker=pending_worker,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "collaborate",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "completed_reports": completed_reports_payload,
                        "pending_worker_index": pending_index,
                    },
                )
                return report, state
            if report.status == "failed":
                return _aggregate_reports(
                    list(zip(self.workers[: pending_index + 1], completed_reports, strict=False))
                    + [(pending_worker, report)],
                    state.run_id,
                    events,
                    "failed",
                ), None
            completed_reports.append(report)
            reports: list[tuple[Worker, Report]] = []
            for worker, rep in zip(
                self.workers[: pending_index + 1],
                completed_reports,
                strict=False,
            ):
                reports.append((worker, rep))
            for worker in self.workers[pending_index + 1 :]:
                rep, next_state = self._run_worker(
                    worker=worker,
                    adapter=adapter,
                    job=root_job,
                    run_id=state.run_id,
                    events=events,
                    emit=emit,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=stream,
                    stream_options=stream_options,
                    structured_output_retries=structured_output_retries,
                    max_iterations=max_iterations,
                    max_tool_calls=max_tool_calls,
                    default_model=default_model,
                    respect_context_window=respect_context_window,
                )
                if next_state is not None:
                    new_payload = {
                        "root_job": root_job.model_dump(mode="json"),
                        "completed_reports": [r.model_dump(mode="json") for _, r in reports],
                        "pending_worker_index": len(reports),
                    }
                    state = _build_workforce_state(
                        state.run_id,
                        "paused",
                        self.name,
                        root_job,
                        next_state,
                        "collaborate",
                        payload=new_payload,
                    )
                    return rep, state
                if rep.status == "failed":
                    return _aggregate_reports(
                        reports + [(worker, rep)],
                        state.run_id,
                        events,
                        "failed",
                    ), None
                reports.append((worker, rep))
            if self.reducer:
                return self.reducer([rep for _, rep in reports]), None
            return _default_reducer(reports, state.run_id, events), None

        report = Report(
            run_id=state.run_id,
            status="failed",
            content=None,
            data=None,
            messages=state.messages,
            tool_calls=state.tool_calls,
            metrics=state.metrics,
            events=events,
            pending_action=None,
            errors=["Unknown workflow stage"],
        )
        return report, None

    async def aresume(
        self,
        *,
        adapter: Any,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: Callable[[str, str, dict[str, Any]], None],
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        default_model: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        payload = state.payload
        stage = payload.get("stage")
        worker_state_payload = payload.get("worker_state")
        if worker_state_payload is None:
            report = Report(
                run_id=state.run_id,
                status="failed",
                content=None,
                data=None,
                messages=state.messages,
                tool_calls=state.tool_calls,
                metrics=state.metrics,
                events=events,
                pending_action=None,
                errors=["Missing worker state"],
            )
            return report, None
        stored_worker_state = RunState.model_validate(worker_state_payload)

        if stage == "manager":
            manager = self.manager or self.workers[0]
            manager_report, manager_state = await self._aresume_worker(
                worker=manager,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if manager_state is not None:
                root_job = _root_job(payload, state.job)
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    manager_state,
                    "manager",
                    payload={"root_job": root_job.model_dump(mode="json")},
                )
                return manager_report, state
            if manager_report.status == "failed":
                return manager_report, None

            root_job = _root_job(payload, state.job)
            selected = _select_worker_name(manager_report, self.workers)
            worker = _find_worker(self.workers, selected)
            report, next_state = await self._arun_worker(
                worker=worker,
                adapter=adapter,
                job=root_job,
                run_id=state.run_id,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "worker",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                return report, state
            if report.status == "failed":
                return report, None
            return report, None

        if stage == "worker":
            root_job = _root_job(payload, state.job)
            worker_name = payload.get("selected_worker")
            worker = _find_worker(self.workers, worker_name)
            report, next_state = await self._aresume_worker(
                worker=worker,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "worker",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "selected_worker": worker.name,
                    },
                )
                return report, state
            if report.status == "failed":
                return report, None
            return report, None

        if stage == "collaborate":
            root_job = _root_job(payload, state.job)
            completed_reports_payload = payload.get("completed_reports", [])
            completed_reports = [Report.model_validate(rep) for rep in completed_reports_payload]
            pending_index = payload.get("pending_worker_index", 0)
            if not isinstance(pending_index, int) or pending_index < 0:
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=events,
                    pending_action=None,
                    errors=["Invalid pending worker index"],
                )
                return report, None
            if pending_index >= len(self.workers):
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=events,
                    pending_action=None,
                    errors=["Invalid pending worker index"],
                )
                return report, None
            pending_worker = self.workers[pending_index]
            report, next_state = await self._aresume_worker(
                worker=pending_worker,
                adapter=adapter,
                state=stored_worker_state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=temperature,
                max_tokens=max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=structured_output_retries,
                max_iterations=max_iterations,
                max_tool_calls=max_tool_calls,
                default_model=default_model,
                respect_context_window=respect_context_window,
            )
            if next_state is not None:
                state = _build_workforce_state(
                    state.run_id,
                    "paused",
                    self.name,
                    root_job,
                    next_state,
                    "collaborate",
                    payload={
                        "root_job": root_job.model_dump(mode="json"),
                        "completed_reports": completed_reports_payload,
                        "pending_worker_index": pending_index,
                    },
                )
                return report, state
            if report.status == "failed":
                return _aggregate_reports(
                    list(zip(self.workers[: pending_index + 1], completed_reports, strict=False))
                    + [(pending_worker, report)],
                    state.run_id,
                    events,
                    "failed",
                ), None
            completed_reports.append(report)
            reports: list[tuple[Worker, Report]] = []
            for worker, rep in zip(
                self.workers[: pending_index + 1],
                completed_reports,
                strict=False,
            ):
                reports.append((worker, rep))
            for worker in self.workers[pending_index + 1 :]:
                rep, next_state = await self._arun_worker(
                    worker=worker,
                    adapter=adapter,
                    job=root_job,
                    run_id=state.run_id,
                    events=events,
                    emit=emit,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    stream=stream,
                    stream_options=stream_options,
                    structured_output_retries=structured_output_retries,
                    max_iterations=max_iterations,
                    max_tool_calls=max_tool_calls,
                    default_model=default_model,
                    respect_context_window=respect_context_window,
                )
                if next_state is not None:
                    new_payload = {
                        "root_job": root_job.model_dump(mode="json"),
                        "completed_reports": [r.model_dump(mode="json") for _, r in reports],
                        "pending_worker_index": len(reports),
                    }
                    state = _build_workforce_state(
                        state.run_id,
                        "paused",
                        self.name,
                        root_job,
                        next_state,
                        "collaborate",
                        payload=new_payload,
                    )
                    return rep, state
                if rep.status == "failed":
                    return _aggregate_reports(
                        reports + [(worker, rep)],
                        state.run_id,
                        events,
                        "failed",
                    ), None
                reports.append((worker, rep))
            if self.reducer:
                return self.reducer([rep for _, rep in reports]), None
            return _default_reducer(reports, state.run_id, events), None

        report = Report(
            run_id=state.run_id,
            status="failed",
            content=None,
            data=None,
            messages=state.messages,
            tool_calls=state.tool_calls,
            metrics=state.metrics,
            events=events,
            pending_action=None,
            errors=["Unknown workflow stage"],
        )
        return report, None



from dataclasses import dataclass, field
from typing import Any

from blackgeorge.core.job import Job
from blackgeorge.core.report import Report


@dataclass
class WorkflowContext:
    job: Job
    outputs: list[Report] = field(default_factory=list)
    artifacts: dict[str, Any] = field(default_factory=dict)



from typing import Literal

MessageRole = Literal["system", "user", "assistant", "tool"]
PendingActionType = Literal["confirmation", "user_input"]
RunStatus = Literal["completed", "paused", "failed", "running"]
WorkforceMode = Literal["managed", "collaborate"]



import asyncio
from typing import Any, cast

from pydantic import BaseModel

from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.report import Report
from blackgeorge.store.state import RunState
from blackgeorge.utils import new_id
from blackgeorge.worker import Worker
from blackgeorge.workflow.context import WorkflowContext
from blackgeorge.workflow.result import StepOutput, StepResult
from blackgeorge.workforce import Workforce


class Flow:
    def __init__(self, desk: Any, steps: list[Any], name: str | None = None) -> None:
        self.desk = desk
        self.steps = steps
        self.name = name or "flow"
        self._run_id = ""
        self._events: list[Event] = []
        self._stream: bool = False

    def emit(self, event_type: str, source: str, payload: dict[str, Any]) -> None:
        if not self._run_id:
            return
        self.desk._emit(self._events, self._run_id, event_type, source, payload)

    def _output_json(self, report: Report) -> Any | None:
        if isinstance(report.data, BaseModel):
            return report.data.model_dump(mode="json", warnings=False)
        return report.data

    async def _run_runner(
        self,
        runner: Worker | Workforce,
        job: Job,
    ) -> tuple[Report, RunState | None]:
        if isinstance(runner, Worker):
            self.desk.register_worker(runner)
            return await runner.arun(
                adapter=self.desk.adapter,
                job=job,
                run_id=self._run_id,
                events=self._events,
                emit=self.emit,
                temperature=self.desk.temperature,
                max_tokens=self.desk.max_tokens,
                stream=self._stream,
                stream_options={"include_usage": True} if self._stream else None,
                structured_output_retries=self.desk.structured_output_retries,
                max_iterations=self.desk.max_iterations,
                max_tool_calls=self.desk.max_tool_calls,
                model_name=runner.model or self.desk.model,
                respect_context_window=self.desk.respect_context_window,
            )
        if isinstance(runner, Workforce):
            self.desk.register_workforce(runner)
            return await runner.arun(
                adapter=self.desk.adapter,
                job=job,
                run_id=self._run_id,
                events=self._events,
                emit=self.emit,
                temperature=self.desk.temperature,
                max_tokens=self.desk.max_tokens,
                stream=self._stream,
                stream_options={"include_usage": True} if self._stream else None,
                structured_output_retries=self.desk.structured_output_retries,
                max_iterations=self.desk.max_iterations,
                max_tool_calls=self.desk.max_tool_calls,
                default_model=self.desk.model,
                respect_context_window=self.desk.respect_context_window,
            )
        raise TypeError("Runner must be Worker or Workforce")

    async def _resume_runner(
        self,
        state: RunState,
        decision_or_input: Any,
        stream: bool,
    ) -> tuple[Report, RunState | None]:
        stream_options = {"include_usage": True} if stream else None
        if state.runner_type == "worker":
            worker = self.desk._workers.get(state.runner_name)
            if worker is None:
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=self._events,
                    pending_action=None,
                    errors=["Worker not registered"],
                )
                return report, None
            worker = cast(Worker, worker)
            return await worker.aresume(
                adapter=self.desk.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=self._events,
                emit=self.emit,
                temperature=self.desk.temperature,
                max_tokens=self.desk.max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=self.desk.structured_output_retries,
                max_iterations=self.desk.max_iterations,
                max_tool_calls=self.desk.max_tool_calls,
                model_name=worker.model or self.desk.model,
                respect_context_window=self.desk.respect_context_window,
            )
        if state.runner_type == "workforce":
            workforce = self.desk._workforces.get(state.runner_name)
            if workforce is None:
                report = Report(
                    run_id=state.run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=state.messages,
                    tool_calls=state.tool_calls,
                    metrics=state.metrics,
                    events=self._events,
                    pending_action=None,
                    errors=["Workforce not registered"],
                )
                return report, None
            workforce = cast(Workforce, workforce)
            return await workforce.aresume(
                adapter=self.desk.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=self._events,
                emit=self.emit,
                temperature=self.desk.temperature,
                max_tokens=self.desk.max_tokens,
                stream=stream,
                stream_options=stream_options,
                structured_output_retries=self.desk.structured_output_retries,
                max_iterations=self.desk.max_iterations,
                max_tool_calls=self.desk.max_tool_calls,
                default_model=self.desk.model,
                respect_context_window=self.desk.respect_context_window,
            )
        report = Report(
            run_id=state.run_id,
            status="failed",
            content=None,
            data=None,
            messages=state.messages,
            tool_calls=state.tool_calls,
            metrics=state.metrics,
            events=self._events,
            pending_action=None,
            errors=["Unknown runner type"],
        )
        return report, None

    def _build_flow_state(
        self,
        job: Job,
        step_index: int,
        reports: list[Report],
        step_state: RunState,
        stream: bool,
    ) -> RunState:
        payload = {
            "step_index": step_index,
            "outputs": [report.model_dump(mode="json") for report in reports],
            "step_state": step_state.model_dump(mode="json"),
            "stream": stream,
        }
        return RunState(
            run_id=self._run_id,
            status="paused",
            runner_type="flow",
            runner_name=self.name,
            job=job,
            messages=step_state.messages,
            tool_calls=step_state.tool_calls,
            pending_action=step_state.pending_action,
            metrics=step_state.metrics,
            iteration=step_state.iteration,
            payload=payload,
        )

    def _restore_outputs(self, payload: Any) -> list[Report]:
        if not isinstance(payload, list):
            return []
        return [Report.model_validate(item) for item in payload]

    def _combine_reports(self, reports: list[Report]) -> Report:
        content_parts: list[str] = []
        for idx, report in enumerate(reports, start=1):
            content_parts.append(f"[step {idx}] {report.content or ''}")
        data = [
            {"index": idx, "content": report.content, "data": report.data}
            for idx, report in enumerate(reports, start=1)
        ]
        return Report(
            run_id=self._run_id,
            status="completed",
            content="\n\n".join(content_parts),
            data=data,
            messages=[message for report in reports for message in report.messages],
            tool_calls=[call for report in reports for call in report.tool_calls],
            metrics={},
            events=self._events,
            pending_action=None,
            errors=[error for report in reports for error in report.errors],
        )

    def _normalize_results(self, results: list[StepOutput]) -> list[StepResult]:
        normalized: list[StepResult] = []
        for item in results:
            if isinstance(item, StepResult):
                normalized.append(item)
            elif isinstance(item, Report):
                normalized.append(StepResult(item, None))
            else:
                raise TypeError("Step must return Report or StepResult")
        return normalized

    async def _run(self, job: Job) -> Report:
        self._run_id = new_id()
        self._events = []
        self._stream = self.desk.stream
        self.desk.run_store.create_run(self._run_id, job.model_dump(mode="json"))
        self.desk.register_flow_run(self._run_id, self)
        self.desk._emit(self._events, self._run_id, "run.started", self.name, {"job_id": job.id})
        context = WorkflowContext(job=job)
        all_reports: list[Report] = []

        for step_index, step in enumerate(self.steps):
            results = await step.execute(self, context)
            for result in self._normalize_results(results):
                report = result.report
                if report.status == "paused":
                    if result.state is None:
                        failed = Report(
                            run_id=self._run_id,
                            status="failed",
                            content=None,
                            data=None,
                            messages=report.messages,
                            tool_calls=report.tool_calls,
                            metrics=report.metrics,
                            events=self._events,
                            pending_action=None,
                            errors=["Missing runner state"],
                        )
                        self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
                        self.desk.run_store.update_run(
                            self._run_id,
                            "failed",
                            None,
                            None,
                            None,
                        )
                        self.desk.unregister_flow_run(self._run_id)
                        return failed
                    self.desk._emit(self._events, self._run_id, "run.paused", self.name, {})
                    result.state.payload["stream"] = self._stream
                    state = self._build_flow_state(
                        job,
                        step_index,
                        all_reports,
                        result.state,
                        self._stream,
                    )
                    self.desk.run_store.update_run(
                        self._run_id,
                        "paused",
                        report.content,
                        None,
                        state,
                    )
                    return report
                if report.status == "failed":
                    self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
                    self.desk.run_store.update_run(
                        self._run_id,
                        "failed",
                        report.content,
                        self._output_json(report),
                        None,
                    )
                    self.desk.unregister_flow_run(self._run_id)
                    return report
                all_reports.append(report)
                context.outputs.append(report)

        if not all_reports:
            empty_report = Report(
                run_id=self._run_id,
                status="failed",
                content=None,
                data=None,
                messages=[],
                tool_calls=[],
                metrics={},
                events=self._events,
                pending_action=None,
                errors=["No steps executed"],
            )
            self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
            self.desk.run_store.update_run(self._run_id, "failed", None, None, None)
            self.desk.unregister_flow_run(self._run_id)
            return empty_report

        if len(all_reports) > 1:
            final_report = self._combine_reports(all_reports)
        else:
            final_report = all_reports[0]
        self.desk._emit(self._events, self._run_id, "run.completed", self.name, {})
        self.desk.run_store.update_run(
            self._run_id,
            "completed",
            final_report.content,
            self._output_json(final_report),
            None,
        )
        self.desk.unregister_flow_run(self._run_id)
        return final_report

    async def _resume(
        self,
        state: RunState,
        decision_or_input: Any,
        stream: bool | None = None,
    ) -> Report:
        payload = state.payload
        step_index = payload.get("step_index")
        step_state_payload = payload.get("step_state")
        if not isinstance(step_index, int) or step_state_payload is None:
            events = self.desk.run_store.get_events(state.run_id)
            failed = Report(
                run_id=state.run_id,
                status="failed",
                content=None,
                data=None,
                messages=state.messages,
                tool_calls=state.tool_calls,
                metrics=state.metrics,
                events=events,
                pending_action=None,
                errors=["Invalid flow state"],
            )
            return failed
        if step_index < 0 or step_index >= len(self.steps):
            events = self.desk.run_store.get_events(state.run_id)
            failed = Report(
                run_id=state.run_id,
                status="failed",
                content=None,
                data=None,
                messages=state.messages,
                tool_calls=state.tool_calls,
                metrics=state.metrics,
                events=events,
                pending_action=None,
                errors=["Invalid step index"],
            )
            return failed

        self._run_id = state.run_id
        self._events = self.desk.run_store.get_events(state.run_id)
        self.desk.register_flow_run(self._run_id, self)
        self.desk._emit(self._events, self._run_id, "run.resumed", self.name, {})

        payload_stream = payload.get("stream")
        if stream is None:
            stream_enabled = (
                payload_stream if isinstance(payload_stream, bool) else self.desk.stream
            )
        else:
            stream_enabled = stream
        self._stream = stream_enabled

        context = WorkflowContext(job=state.job)
        all_reports = self._restore_outputs(payload.get("outputs"))
        context.outputs.extend(all_reports)

        step_state = RunState.model_validate(step_state_payload)
        step_state.payload["stream"] = stream_enabled
        report, updated_state = await self._resume_runner(
            step_state,
            decision_or_input,
            stream_enabled,
        )
        if report.status == "paused":
            if updated_state is None:
                failed = Report(
                    run_id=self._run_id,
                    status="failed",
                    content=None,
                    data=None,
                    messages=report.messages,
                    tool_calls=report.tool_calls,
                    metrics=report.metrics,
                    events=self._events,
                    pending_action=None,
                    errors=["Missing runner state"],
                )
                self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
                self.desk.run_store.update_run(self._run_id, "failed", None, None, None)
                self.desk.unregister_flow_run(self._run_id)
                return failed
            self.desk._emit(self._events, self._run_id, "run.paused", self.name, {})
            updated_state.payload["stream"] = stream_enabled
            flow_state = self._build_flow_state(
                state.job,
                step_index,
                all_reports,
                updated_state,
                stream_enabled,
            )
            self.desk.run_store.update_run(
                self._run_id,
                "paused",
                report.content,
                None,
                flow_state,
            )
            return report
        if report.status == "failed":
            self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
            self.desk.run_store.update_run(
                self._run_id,
                "failed",
                report.content,
                self._output_json(report),
                None,
            )
            self.desk.unregister_flow_run(self._run_id)
            return report

        all_reports.append(report)
        context.outputs.append(report)

        for next_index, step in enumerate(self.steps[step_index + 1 :], start=step_index + 1):
            results = await step.execute(self, context)
            for result in self._normalize_results(results):
                report = result.report
                if report.status == "paused":
                    if result.state is None:
                        failed = Report(
                            run_id=self._run_id,
                            status="failed",
                            content=None,
                            data=None,
                            messages=report.messages,
                            tool_calls=report.tool_calls,
                            metrics=report.metrics,
                            events=self._events,
                            pending_action=None,
                            errors=["Missing runner state"],
                        )
                        self.desk._emit(
                            self._events,
                            self._run_id,
                            "run.failed",
                            self.name,
                            {},
                        )
                        self.desk.run_store.update_run(
                            self._run_id,
                            "failed",
                            None,
                            None,
                            None,
                        )
                        self.desk.unregister_flow_run(self._run_id)
                        return failed
                    self.desk._emit(self._events, self._run_id, "run.paused", self.name, {})
                    result.state.payload["stream"] = stream_enabled
                    flow_state = self._build_flow_state(
                        state.job,
                        next_index,
                        all_reports,
                        result.state,
                        stream_enabled,
                    )
                    self.desk.run_store.update_run(
                        self._run_id,
                        "paused",
                        report.content,
                        None,
                        flow_state,
                    )
                    return report
                if report.status == "failed":
                    self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
                    self.desk.run_store.update_run(
                        self._run_id,
                        "failed",
                        report.content,
                        self._output_json(report),
                        None,
                    )
                    self.desk.unregister_flow_run(self._run_id)
                    return report
                all_reports.append(report)
                context.outputs.append(report)

        if not all_reports:
            empty_report = Report(
                run_id=self._run_id,
                status="failed",
                content=None,
                data=None,
                messages=[],
                tool_calls=[],
                metrics={},
                events=self._events,
                pending_action=None,
                errors=["No steps executed"],
            )
            self.desk._emit(self._events, self._run_id, "run.failed", self.name, {})
            self.desk.run_store.update_run(self._run_id, "failed", None, None, None)
            self.desk.unregister_flow_run(self._run_id)
            return empty_report

        if len(all_reports) > 1:
            final_report = self._combine_reports(all_reports)
        else:
            final_report = all_reports[0]
        self.desk._emit(self._events, self._run_id, "run.completed", self.name, {})
        self.desk.run_store.update_run(
            self._run_id,
            "completed",
            final_report.content,
            self._output_json(final_report),
            None,
        )
        self.desk.unregister_flow_run(self._run_id)
        return final_report

    async def arun(self, job: Job) -> Report:
        return await self._run(job)

    async def aresume(
        self,
        report: Report,
        decision_or_input: Any,
        *,
        stream: bool | None = None,
    ) -> Report:
        record = self.desk.run_store.get_run(report.run_id)
        if record is None or record.state is None:
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=report.events,
                pending_action=None,
                errors=["No stored state"],
            )
            return failed
        if record.state.runner_type != "flow":
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=report.events,
                pending_action=None,
                errors=["Run is not a flow"],
            )
            return failed
        return await self._resume(record.state, decision_or_input, stream=stream)

    def run(self, job: Job) -> Report:
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return asyncio.run(self._run(job))
        raise RuntimeError(
            "Flow.run cannot be called from a running event loop. Use Flow.arun instead."
        )

    def resume(
        self,
        report: Report,
        decision_or_input: Any,
        *,
        stream: bool | None = None,
    ) -> Report:
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return asyncio.run(self.aresume(report, decision_or_input, stream=stream))
        raise RuntimeError(
            "Flow.resume cannot be called from a running event loop. Use Flow.aresume instead."
        )



import asyncio
import concurrent.futures
import contextlib
import json
import time
from inspect import isawaitable, iscoroutinefunction
from typing import Any

from pydantic import BaseModel

from blackgeorge.async_utils import run_coroutine_in_thread, run_coroutine_sync
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.tools.base import ProgressCallback, Tool, ToolResult


def _to_content(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, BaseModel):
        return value.model_dump_json()
    try:
        return json.dumps(value, ensure_ascii=True, default=str)
    except TypeError:
        return str(value)


def _run_coroutine_in_thread(coro: Any) -> Any:
    return run_coroutine_in_thread(coro)


def _run_coroutine_sync(coro: Any) -> Any:
    return run_coroutine_sync(coro)


def _run_sync_call(tool: Tool, args: dict[str, Any]) -> Any:
    if iscoroutinefunction(tool.callable):
        return _run_coroutine_sync(tool.callable(**args))
    return tool.callable(**args)


def _validate_tool_call(call: ToolCall, tool: Tool) -> dict[str, Any]:
    validated = tool.input_model.model_validate(call.arguments)
    return validated.model_dump()


def _sync_invoke_hook(hook: Any, *args: Any) -> None:
    result = hook(*args)
    if isawaitable(result):
        _run_coroutine_sync(result)


def _sync_pre_hooks(tool: Tool, call: ToolCall) -> None:
    for pre_hook in tool.pre:
        _sync_invoke_hook(pre_hook, call)


def _sync_post_hooks(tool: Tool, call: ToolCall, result: ToolResult) -> None:
    for post_hook in tool.post:
        _sync_invoke_hook(post_hook, call, result)


async def _async_invoke_hook(hook: Any, *args: Any) -> None:
    result = hook(*args)
    if isawaitable(result):
        await result


async def _async_pre_hooks(tool: Tool, call: ToolCall) -> None:
    for pre_hook in tool.pre:
        await _async_invoke_hook(pre_hook, call)


async def _async_post_hooks(tool: Tool, call: ToolCall, result: ToolResult) -> None:
    for post_hook in tool.post:
        await _async_invoke_hook(post_hook, call, result)


def _backoff_delay(retry_delay: float, attempt: int) -> float:
    return float(retry_delay * (2**attempt))


def _execute_sync_with_retries(
    tool: Tool,
    args: dict[str, Any],
) -> ToolResult:
    retries = tool.retries
    retry_delay = tool.retry_delay
    last_result: ToolResult | None = None

    for attempt in range(retries + 1):
        last_result = _execute_sync_once(tool, args, tool.timeout)
        if last_result.error is None:
            break
        if last_result.cancelled:
            break
        if attempt < retries:
            time.sleep(_backoff_delay(retry_delay, attempt))

    return last_result or ToolResult(error="No execution result")


def _execute_sync_once(tool: Tool, args: dict[str, Any], timeout: float | None) -> ToolResult:
    try:
        if timeout is None:
            result = _run_sync_call(tool, args)
        else:
            with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(_run_sync_call, tool, args)
                try:
                    result = future.result(timeout=timeout)
                except concurrent.futures.TimeoutError:
                    future.cancel()
                    return ToolResult(
                        error=f"Tool execution timed out after {timeout}s",
                        timed_out=True,
                    )
        if isinstance(result, ToolResult):
            return result
        return ToolResult(content=_to_content(result), data=result)
    except Exception as exc:
        return ToolResult(error=str(exc))


def execute_tool(tool: Tool, call: ToolCall) -> ToolResult:
    _sync_pre_hooks(tool, call)

    try:
        args = _validate_tool_call(call, tool)
    except Exception as exc:
        tool_result = ToolResult(error=str(exc))
        _sync_post_hooks(tool, call, tool_result)
        return tool_result

    tool_result = _execute_sync_with_retries(tool, args)
    _sync_post_hooks(tool, call, tool_result)

    return tool_result


async def _execute_once(
    tool: Tool,
    args: dict[str, Any],
    timeout: float | None,
    cancel_event: asyncio.Event | None,
) -> ToolResult:
    async def run_callable() -> Any:
        if iscoroutinefunction(tool.callable):
            return await tool.callable(**args)
        return await asyncio.to_thread(tool.callable, **args)

    if cancel_event is not None and cancel_event.is_set():
        return ToolResult(error="Cancelled before execution", cancelled=True)

    task = asyncio.create_task(run_callable())
    cancel_task: asyncio.Task[bool] | None = None
    tasks: set[asyncio.Task[Any]] = {task}
    if cancel_event is not None:
        cancel_task = asyncio.create_task(cancel_event.wait())
        tasks.add(cancel_task)

    try:
        done, _ = await asyncio.wait(
            tasks,
            timeout=timeout,
            return_when=asyncio.FIRST_COMPLETED,
        )

        if task in done:
            result = await task
            if isinstance(result, ToolResult):
                return result
            return ToolResult(content=_to_content(result), data=result)

        if cancel_task is not None and cancel_task in done:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task
            return ToolResult(error="Tool execution cancelled", cancelled=True)

        task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await task
        return ToolResult(error=f"Tool execution timed out after {timeout}s", timed_out=True)
    except asyncio.CancelledError:
        task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await task
        return ToolResult(error="Tool execution cancelled", cancelled=True)
    except Exception as exc:
        return ToolResult(error=str(exc))
    finally:
        if cancel_task is not None and not cancel_task.done():
            cancel_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await cancel_task


async def aexecute_tool(
    tool: Tool,
    call: ToolCall,
    *,
    cancel_event: asyncio.Event | None = None,
    on_progress: ProgressCallback | None = None,
) -> ToolResult:
    await _async_pre_hooks(tool, call)

    try:
        args = _validate_tool_call(call, tool)
    except Exception as exc:
        tool_result = ToolResult(error=f"Validation error: {exc}")
        await _async_post_hooks(tool, call, tool_result)
        return tool_result

    retries = tool.retries
    retry_delay = tool.retry_delay
    last_result: ToolResult | None = None

    for attempt in range(retries + 1):
        if cancel_event is not None and cancel_event.is_set():
            return ToolResult(error="Cancelled", cancelled=True)

        if on_progress is not None and attempt > 0:
            on_progress(f"Retry attempt {attempt}/{retries}")

        last_result = await _execute_once(tool, args, tool.timeout, cancel_event)

        if last_result.error is None:
            break

        if last_result.cancelled:
            break

        if attempt < retries:
            delay = _backoff_delay(retry_delay, attempt)
            if on_progress is not None:
                on_progress(f"Waiting {delay:.1f}s before retry")
            await asyncio.sleep(delay)

    tool_result = last_result or ToolResult(error="No execution result")

    await _async_post_hooks(tool, call, tool_result)

    return tool_result



import asyncio
import concurrent.futures
import threading
from collections.abc import Coroutine
from typing import Any


def run_coroutine_in_thread[T](coro: Coroutine[Any, Any, T]) -> T:
    future: concurrent.futures.Future[T] = concurrent.futures.Future()

    def runner() -> None:
        try:
            result = asyncio.run(coro)
            future.set_result(result)
        except BaseException as exc:
            future.set_exception(exc)

    thread = threading.Thread(target=runner)
    thread.start()
    thread.join()
    return future.result()


def run_coroutine_sync[T](coro: Coroutine[Any, Any, T]) -> T:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(coro)
    return run_coroutine_in_thread(coro)
